from .. import excepttypes, objtypes
from ..constraints import ObjectConstraint, maybeThrow, throw

from .base import BaseOp

class Throw(BaseOp):
    def __init__(self, parent, args):
        super(Throw, self).__init__(parent, args, makeException=True)
        self.env = parent.env

    def propagateConstraints(self, x):
        if x.null:
            t = x.types
            exact = list(t.exact) + [excepttypes.NullPtr]
            return throw(ObjectConstraint.fromTops(t.env, t.supers, exact, nonnull=True))
        return throw(x)

# Dummy instruction that can throw anything
class MagicThrow(BaseOp):
    def __init__(self, parent):
        super(MagicThrow, self).__init__(parent, [], makeException=True)
        self.eout = ObjectConstraint.fromTops(parent.env, [objtypes.ThrowableTT], [], nonnull=True)

    def propagateConstraints(self):
        return maybeThrow(self.eout)



from .base import BaseOp

class Convert(BaseOp):
    def __init__(self, parent, arg, source_ssa, target_ssa):
        super(Convert, self).__init__(parent, [arg])
        self.source = source_ssa
        self.target = target_ssa
        self.rval = parent.makeVariable(target_ssa, origin=self)



from __future__ import print_function

import collections
import re
import sys

from . import token_regexes as res

class AsssemblerError(Exception):
    pass

Token = collections.namedtuple('Token', 'type val pos')

TOKENS = [
    ('WHITESPACE', r'[ \t]+'),
    ('WORD', res.WORD + res.FOLLOWED_BY_WHITESPACE),
    ('DIRECTIVE', res.DIRECTIVE + res.FOLLOWED_BY_WHITESPACE),
    ('LABEL_DEF', res.LABEL_DEF + res.FOLLOWED_BY_WHITESPACE),
    ('NEWLINES', res.NEWLINES),
    ('REF', res.REF + res.FOLLOWED_BY_WHITESPACE),
    ('COLON', r':' + res.FOLLOWED_BY_WHITESPACE),
    ('EQUALS', r'=' + res.FOLLOWED_BY_WHITESPACE),
    ('INT_LITERAL', res.INT_LITERAL + res.FOLLOWED_BY_WHITESPACE),
    ('DOUBLE_LITERAL', res.FLOAT_LITERAL + res.FOLLOWED_BY_WHITESPACE),
    ('STRING_LITERAL', res.STRING_LITERAL + res.FOLLOWED_BY_WHITESPACE),
]
REGEX = re.compile('|'.join('(?P<{}>{})'.format(*pair) for pair in TOKENS), re.VERBOSE)
# For error detection
STRING_START_REGEX = re.compile(res.STRING_START)
WORD_LIKE_REGEX = re.compile(r'.\S*')

MAXLINELEN = 80
def formatError(source, filename, message, point, point2):
    try:
        start = source.rindex('\n', 0, point) + 1
    except ValueError:
        start = 0
    line_start = start

    try:
        end = source.index('\n', start) + 1
    except ValueError:    # pragma: no cover
        end = len(source) + 1

    # Find an 80 char section of the line around the point of interest to display
    temp = min(point2, point + MAXLINELEN//2)
    if temp < start + MAXLINELEN:
        end = min(end, start + MAXLINELEN)
    elif point >= end - MAXLINELEN:
        start = max(start, end - MAXLINELEN)
    else:
        mid = (point + temp) // 2
        start = max(start, mid - MAXLINELEN//2)
        end = min(end, start + MAXLINELEN)
    point2 = min(point2, end)
    assert line_start <= start <= point < point2 <= end

    pchars = [' '] * (end - start)
    for i in range(point - start, point2 - start):
        pchars[i] = '~'
    pchars[point - start] = '^'
    lineno = source[:line_start].count('\n') + 1
    colno = point - line_start + 1
    return '{}:{}:{}: {}\n{}\n{}'.format(filename, lineno, colno,
        message, source[start:end].rstrip('\n'), ''.join(pchars))

class Tokenizer(object):
    def __init__(self, source, filename):
        self.s = source
        self.pos = 0
        self.atlineend = True
        if isinstance(filename, bytes):
            filename = filename.decode()
        self.filename = filename.rpartition('/')[-1]

    def error(self, error, *notes):
        printerr = lambda s: print(s, file=sys.stderr)
        message, point, point2 = error
        printerr(formatError(self.s, self.filename, 'error: ' + message, point, point2))
        for message, point, point2 in notes:
            printerr(formatError(self.s, self.filename, 'note: ' + message, point, point2))
        raise AsssemblerError()

    def _nextsub(self):
        match = REGEX.match(self.s, self.pos)
        if match is None:
            if self.atend():
                return Token('EOF', '', self.pos)
            else:
                str_match = STRING_START_REGEX.match(self.s, self.pos)
                if str_match is not None:
                    self.error(('Invalid escape sequence or character in string literal', str_match.end(), str_match.end()+1))

                match = WORD_LIKE_REGEX.match(self.s, self.pos)
                return Token('INVALID_TOKEN', match.group(), match.start())
        assert match.start() == match.pos == self.pos

        self.pos = match.end()
        return Token(match.lastgroup, match.group(), match.start())

    def next(self):
        tok = self._nextsub()
        while tok.type == 'WHITESPACE' or self.atlineend and tok.type == 'NEWLINES':
            tok = self._nextsub()
        self.atlineend = tok.type == 'NEWLINES'

        if tok.type == 'INT_LITERAL' and tok.val.lower().endswith('l'):
            return tok._replace(type='LONG_LITERAL')
        elif tok.type == 'DOUBLE_LITERAL' and tok.val.lower().endswith('f'):
            return tok._replace(type='FLOAT_LITERAL')
        return tok

    def atend(self): return self.pos == len(self.s)



from .base import BaseJump

class Return(BaseJump):
    def __init__(self, parent, arguments):
        super(Return, self).__init__(parent, arguments)

class Rethrow(BaseJump):
    def __init__(self, parent, arguments):
        super(Rethrow, self).__init__(parent, arguments)



import struct

from ..ssa import objtypes
from ..verifier.descriptors import parseFieldDescriptor

from . import ast, ast2, javamethod, throws
from .reserved import reserved_identifiers

def loadConstValue(cpool, index):
    entry_type = cpool.pool[index][0]
    args = cpool.getArgs(index)

    # Note: field constant values cannot be class literals
    tt = {'Int':objtypes.IntTT, 'Long':objtypes.LongTT,
        'Float':objtypes.FloatTT, 'Double':objtypes.DoubleTT,
        'String':objtypes.StringTT}[entry_type]
    return ast.Literal(tt, args[0]).fixLiterals()

def _getField(field):
    flags = [x.lower() for x in sorted(field.flags) if x not in ('SYNTHETIC','ENUM')]
    desc = field.descriptor
    dtype = objtypes.verifierToSynthetic(parseFieldDescriptor(desc, unsynthesize=False)[0])

    initexpr = None
    if field.static:
        cpool = field.class_.cpool
        const_attrs = [data for name,data in field.attributes if name == 'ConstantValue']
        if const_attrs:
            assert len(const_attrs) == 1
            data = const_attrs[0]
            index = struct.unpack('>h', data)[0]
            initexpr = loadConstValue(cpool, index)
    return ast2.FieldDef(' '.join(flags), ast.TypeName(dtype), field.class_, field.name, desc, initexpr)

def _getMethod(method, cb, forbidden_identifiers, skip_errors):
    try:
        graph = cb(method) if method.code is not None else None
        print('Decompiling method', method.name.encode('utf8'), method.descriptor.encode('utf8'))
        code_ast = javamethod.generateAST(method, graph, forbidden_identifiers)
        return code_ast
    except Exception as e:
        if not skip_errors:
            raise

        import traceback
        message = traceback.format_exc()
        code_ast = javamethod.generateAST(method, None, forbidden_identifiers)
        code_ast.comments.add(message)
        print(message)
        return code_ast

# Method argument allows decompilng only a single method, primarily useful for debugging
def generateAST(cls, cb, skip_errors, method=None, add_throws=False):
    methods = cls.methods if method is None else [cls.methods[method]]
    fi = set(reserved_identifiers)
    for field in cls.fields:
        fi.add(field.name)
    forbidden_identifiers = frozenset(fi)

    myflags = [x.lower() for x in sorted(cls.flags) if x not in ('INTERFACE','SUPER','SYNTHETIC','ANNOTATION','ENUM')]
    isInterface = 'INTERFACE' in cls.flags

    superc = cls.supername
    interfaces = [cls.cpool.getArgsCheck('Class', index) for index in cls.interfaces_raw] # todo - change when class actually loads interfaces

    field_defs = [_getField(f) for f in cls.fields]
    method_defs = [_getMethod(m, cb, forbidden_identifiers, skip_errors) for m in methods]
    if add_throws:
        throws.addSingle(cls.env, method_defs)
    return ast2.ClassDef(' '.join(myflags), isInterface, cls.name, superc, interfaces, field_defs, method_defs)



from .. import objtypes
from ..constraints import ObjectConstraint
from ..exceptionset import CatchSetManager, ExceptionSet

from .base import BaseJump
from .goto import Goto

class OnException(BaseJump):
    def __init__(self, parent, throwvar, chpairs, fallthrough=None):
        super(OnException, self).__init__(parent, [throwvar])
        self.default = fallthrough
        self.cs = CatchSetManager.new(parent.env, chpairs)
        self.cs.pruneKeys()

    def replaceExceptTarget(self, old, new):
        self.cs.replaceKeys({old:new})

    def replaceNormalTarget(self, old, new):
        self.default = new if self.default == old else self.default

    def replaceBlocks(self, blockDict):
        self.cs.replaceKeys(blockDict)
        if self.default is not None:
            self.default = blockDict.get(self.default, self.default)

    def reduceSuccessors(self, pairsToRemove):
        for (child, t) in pairsToRemove:
            if t:
                self.cs.mask -= self.cs.sets[child]
                del self.cs.sets[child]
            else:
                self.replaceNormalTarget(child, None)

        self.cs.pruneKeys()
        if not self.cs.sets:
            if not self.default:
                return None
            return Goto(self.parent, self.default)
        return self

    def getNormalSuccessors(self):
        return [self.default] if self.default is not None else []

    def getExceptSuccessors(self):
        return self.cs.sets.keys()

    def clone(self):
        new = super(OnException, self).clone()
        new.cs = self.cs.copy()
        return new

    ###############################################################################
    def constrainJumps(self, x):
        if x is None:
            mask = ExceptionSet.EMPTY
        else:
            mask = ExceptionSet(x.types.env, [(objtypes.className(tt),()) for tt in x.types.supers | x.types.exact])
        self.cs.newMask(mask)
        return self.reduceSuccessors([])



from .base import BaseJump

class Goto(BaseJump):
    def __init__(self, parent, target):
        super(Goto, self).__init__(parent, [])
        self.successors = [target]

    def replaceBlocks(self, blockDict):
        self.successors = [blockDict.get(key,key) for key in self.successors]

    def getNormalSuccessors(self):
        return self.successors

    def reduceSuccessors(self, pairsToRemove):
        if (self.successors[0], False) in pairsToRemove:
            return None
        return self



import itertools

from .. import bytecode, error as error_types, opnames as ops

from .descriptors import parseFieldDescriptor, parseMethodDescriptor, parseUnboundMethodDescriptor
from .verifier_types import OBJECT_INFO, T_ADDRESS, T_ARRAY, T_DOUBLE, T_FLOAT, T_INT, T_INT_CONST, T_INVALID, T_LONG, T_NULL, T_OBJECT, T_UNINIT_OBJECT, T_UNINIT_THIS, decrementDim, exactArrayFrom, fullinfo_t, mergeTypes

class VerifierTypesState(object):
    def __init__(self, stack, locals, masks):
        self.stack = stack
        self.locals = locals
        self.masks = masks

    def copy(self): return VerifierTypesState(self.stack, self.locals, self.masks)

    def withExcept(self, t): return VerifierTypesState([t], self.locals, self.masks)

    def pop(self, n):
        if n == 0:
            return []
        self.stack, popped = self.stack[:-n], self.stack[-n:]
        return popped

    def push(self, vals):
        self.stack = self.stack + list(vals)

    def setLocal(self, i, v):
        if len(self.locals) < i:
            self.locals = self.locals + [T_INVALID]*(i - len(self.locals))
        self.locals = self.locals[:i] + [v] + self.locals[i+1:]

        new = frozenset([i])
        self.masks = [(addr, old | new) for addr, old in self.masks]

    def local(self, i):
        if len(self.locals) <= i:
            return T_INVALID
        return self.locals[i]

    def jsr(self, target):
        self.masks = self.masks + [(target, frozenset())]

    def replace(self, old, new):
        self.stack = [(new if v == old else v) for v in self.stack]

        mask = frozenset(i for i, v in enumerate(self.locals) if v == old)
        self.locals = [(new if v == old else v) for v in self.locals]
        self.masks = [(addr, oldmask | mask) for addr, oldmask in self.masks]

    def invalidateNews(self):
        # Doesn't need to update mask
        self.stack = [(T_INVALID if v.tag == '.new' else v) for v in self.stack]
        self.locals = [(T_INVALID if v.tag == '.new' else v) for v in self.locals]

    def maskFor(self, called):
        self.masks = self.masks[:]
        target, mask = self.masks.pop()
        while target != called:
            target, mask = self.masks.pop()
        return mask

    def returnTo(self, called, jsrstate):
        mask = self.maskFor(called)
        # merge locals using mask
        zipped = itertools.izip_longest(self.locals, jsrstate.locals, fillvalue=T_INVALID)
        self.locals = [(x if i in mask else y) for i,(x,y) in enumerate(zipped)]

    def merge(self, other, env):
        old_triple = self.stack, self.locals, self.masks
        assert len(self.stack) == len(other.stack)
        self.stack = [mergeTypes(env, new, old) for old, new in zip(self.stack, other.stack)]
        self.locals = [mergeTypes(env, new, old) for old, new in zip(self.locals, other.locals)]
        while self.locals and self.locals[-1] == T_INVALID:
            self.locals.pop()

        # Merge Masks
        last_match = -1
        mergedmasks = []
        for entry1, mask1 in self.masks:
            for j, (entry2, mask2) in enumerate(other.masks):
                if j > last_match and entry1 == entry2:
                    item = entry1, (mask1 | mask2)
                    mergedmasks.append(item)
                    last_match = j
        self.masks = mergedmasks
        return (self.stack, self.locals, self.masks) != old_triple

def stateFromInitialArgs(args): return VerifierTypesState([], args[:], [])



_invoke_ops = (ops.INVOKESPECIAL, ops.INVOKESTATIC, ops.INVOKEVIRTUAL, ops.INVOKEINTERFACE, ops.INVOKEINIT, ops.INVOKEDYNAMIC)

def _loadFieldDesc(cpool, ind):
    target, name, desc = cpool.getArgsCheck('Field', ind)
    return parseFieldDescriptor(desc)

def _loadMethodDesc(cpool, ind):
    target, name, desc = cpool.getArgs(ind)
    return parseMethodDescriptor(desc)

def _indexToCFMInfo(cpool, ind, typen):
    actual = cpool.getType(ind)
    # JVM_GetCPMethodClassNameUTF accepts both
    assert actual == typen or actual == 'InterfaceMethod' and typen == 'Method'

    cname = cpool.getArgs(ind)[0]
    if cname.startswith('[') or cname.endswith(';'):
        try:
            return parseFieldDescriptor(cname)[0]
        except ValueError as e:
            return T_INVALID
    else:
        return T_OBJECT(cname)

# Instructions which pop a fixed amount
_popAmount = {
    ops.ARRLOAD_OBJ: 2,
    ops.ARRSTORE_OBJ: 3,
    ops.ARRLOAD: 2,
    ops.TRUNCATE: 1,
    ops.LCMP: 4,
    ops.IF_A: 1,
    ops.IF_I: 1,
    ops.IF_ACMP: 2,
    ops.IF_ICMP: 2,
    ops.SWITCH: 1,
    ops.NEWARRAY: 1,
    ops.ANEWARRAY: 1,
    ops.ARRLEN: 1,
    ops.THROW: 1,
    ops.CHECKCAST: 1,
    ops.INSTANCEOF: 1,
    ops.MONENTER: 1,
    ops.MONEXIT: 1,
    ops.GETFIELD: 1,

    ops.NOP: 0,
    ops.CONSTNULL: 0,
    ops.CONST: 0,
    ops.LDC: 0,
    ops.LOAD: 0,
    ops.IINC: 0,
    ops.GOTO: 0,
    ops.JSR: 0,
    ops.RET: 0,
    ops.NEW: 0,
    ops.GETSTATIC: 0,
}
# Instructions which pop a variable amount depending on whether type is category 2
_popAmountVar = {
    ops.STORE: (1, 0),
    ops.NEG: (1, 0),
    ops.CONVERT: (1, 0),

    ops.ADD: (2, 0),
    ops.SUB: (2, 0),
    ops.MUL: (2, 0),
    ops.DIV: (2, 0),
    ops.REM: (2, 0),
    ops.XOR: (2, 0),
    ops.OR: (2, 0),
    ops.AND: (2, 0),
    ops.FCMP: (2, 0),

    ops.SHL: (1, 1),
    ops.SHR: (1, 1),
    ops.USHR: (1, 1),

    ops.ARRSTORE: (1, 2),
}
# Generic stack codes
genericStackCodes = {
    ops.POP: (1, []),
    ops.POP2: (2, []),
    ops.DUP: (1, [0, 0]),
    ops.DUPX1: (2, [1, 0, 1]),
    ops.DUPX2: (3, [2, 0, 1, 2]),
    ops.DUP2: (2, [0, 1, 0, 1]),
    ops.DUP2X1: (3, [1, 2, 0, 1, 2]),
    ops.DUP2X2: (4, [2, 3, 0, 1, 2, 3]),
    ops.SWAP: (2, [1, 0]),
}

def _getPopAmount(cpool, instr, method):
    op = instr[0]
    if op in _popAmount:
        return _popAmount[op]
    if op in _popAmountVar:
        a, b = _popAmountVar[op]
        cat = 2 if instr[1] in 'JD' else 1
        return a * cat + b
    if op in genericStackCodes:
        return genericStackCodes[op][0]

    if op == ops.MULTINEWARRAY:
        return instr[2]
    elif op == ops.RETURN:
        return len(parseMethodDescriptor(method.descriptor)[1])
    elif op in (ops.PUTFIELD, ops.PUTSTATIC):
        args = len(_loadFieldDesc(cpool, instr[1]))
        if op == ops.PUTFIELD:
            args += 1
        return args
    elif op in _invoke_ops:
        args = len(_loadMethodDesc(cpool, instr[1])[0])
        if op != ops.INVOKESTATIC and op != ops.INVOKEDYNAMIC:
            args += 1
        return args

codes = dict(zip('IFJD', [T_INT, T_FLOAT, T_LONG, T_DOUBLE]))
def _getStackResult(cpool, instr, key):
    op = instr[0]

    if op in (ops.TRUNCATE, ops.LCMP, ops.FCMP, ops.ARRLEN, ops.INSTANCEOF):
        return T_INT
    elif op in (ops.ADD, ops.SUB, ops.MUL, ops.DIV, ops.REM, ops.XOR, ops.AND, ops.OR, ops.SHL, ops.SHR, ops.USHR, ops.NEG):
        return codes[instr[1]]
    elif op == ops.CONSTNULL:
        return T_NULL
    elif op == ops.CONST:
        if instr[1] == 'I':
            return T_INT_CONST(instr[2])
        return codes[instr[1]]
    elif op == ops.ARRLOAD:
        return codes.get(instr[1], T_INT)
    elif op == ops.CONVERT:
        return codes[instr[2]]

    elif op == ops.LDC:
        return {
            'Int': T_INT,
            'Long': T_LONG,
            'Float': T_FLOAT,
            'Double': T_DOUBLE,
            'String': T_OBJECT('java/lang/String'),
            'Class': T_OBJECT('java/lang/Class'),
            'MethodType': T_OBJECT('java/lang/invoke/MethodType'),
            'MethodHandle': T_OBJECT('java/lang/invoke/MethodHandle'),
        }[cpool.getType(instr[1])]

    elif op == ops.JSR:
        return T_ADDRESS(instr[1])

    elif op in (ops.CHECKCAST, ops.NEW, ops.ANEWARRAY, ops.MULTINEWARRAY):
        target = _indexToCFMInfo(cpool, instr[1], 'Class')
        if op == ops.ANEWARRAY:
            return T_ARRAY(target)
        elif op == ops.NEW:
            return T_UNINIT_OBJECT(key)
        return target
    elif op == ops.NEWARRAY:
        return parseFieldDescriptor('[' + instr[1])[0]

    elif op in (ops.GETFIELD, ops.GETSTATIC):
        return _loadFieldDesc(cpool, instr[1])[0]
    elif op in _invoke_ops:
        out = _loadMethodDesc(cpool, instr[1])[1]
        assert 0 <= len(out) <= 2
        return out[0] if out else None

class InstructionNode(object):
    __slots__ = "key code env class_ cpool instruction op visited changed offsetToIndex indexToOffset state out_state jsrTarget next_instruction returnedFrom successors pop_amount stack_push stack_code target_type isThisCtor".split()

    def __init__(self, code, offsetToIndex, indexToOffset, key):
        self.key = key
        assert(self.key is not None) # if it is this will cause problems with origin tracking

        self.code = code
        self.env = code.class_.env
        self.class_ = code.class_
        self.cpool = self.class_.cpool

        self.instruction = code.bytecode[key]
        self.op = self.instruction[0]

        self.visited, self.changed = False, False
        # store for usage calculating JSRs, finding successor instructions and the like
        self.offsetToIndex = offsetToIndex
        self.indexToOffset = indexToOffset

        self.state = None

        # Fields to be assigned later
        self.jsrTarget = None
        self.next_instruction = None
        self.returnedFrom = None
        self.successors = None

        self.pop_amount = -1
        self.stack_push = []
        self.stack_code = None

        # for blockmaker
        self.target_type = None
        self.isThisCtor = False
        self.out_state = None # store out state for JSR/RET instructions

        self._precomputeValues()

    def _removeInterface(self, vt):
        if vt.tag == '.obj' and vt.extra is not None and self.env.isInterface(vt.extra, forceCheck=True):
            return T_ARRAY(OBJECT_INFO, vt.dim)
        return vt

    def _precomputeValues(self):
        # parsed_desc, successors
        off_i = self.offsetToIndex[self.key]
        self.next_instruction = self.indexToOffset[off_i+1] # None if end of code
        op = self.instruction[0]

        self.pop_amount = _getPopAmount(self.cpool, self.instruction, self.code.method)

        # cache these, since they're not state dependent
        result = _getStackResult(self.cpool, self.instruction, self.key)
        # temporary hack
        if op == ops.CHECKCAST:
            result = self._removeInterface(result)

        if result is not None:
            self.stack_push = [result]
            if result in (T_LONG, T_DOUBLE):
                self.stack_push.append(T_INVALID)

        if op in genericStackCodes:
            self.stack_code = genericStackCodes[op][1]

        if op == ops.NEW:
            self.target_type = _indexToCFMInfo(self.cpool, self.instruction[1], 'Class')

        # Now get successors
        next_ = self.next_instruction
        if op in (ops.IF_A, ops.IF_I, ops.IF_ICMP, ops.IF_ACMP):
            self.successors = next_, self.instruction[2]
        elif op in (ops.JSR, ops.GOTO):
            self.successors = self.instruction[1],
        elif op in (ops.RETURN, ops.THROW):
            self.successors = ()
        elif op == ops.RET:
            self.successors = None # calculate it when the node is reached
        elif op == ops.SWITCH:
            opname, default, jumps = self.instruction
            targets = (default,)
            if jumps:
                targets += zip(*jumps)[1]
            self.successors = targets
        else:
            self.successors = next_,

    def _getNewState(self, iNodes):
        state = self.state.copy()
        popped = state.pop(self.pop_amount)

        # Local updates/reading
        op = self.instruction[0]
        if op == ops.LOAD:
            state.push([state.local(self.instruction[2])])
            if self.instruction[1] in 'JD':
                state.push([T_INVALID])
        elif op == ops.STORE:
            for i, val in enumerate(popped):
                state.setLocal(self.instruction[2] + i, val)
        elif op == ops.IINC:
            state.setLocal(self.instruction[1], T_INT) # Make sure to clobber constants
        elif op == ops.JSR:
            state.jsr(self.instruction[1])
        elif op == ops.NEW:
            # This should never happen, but better safe than sorry.
            state.replace(self.stack_push[0], T_INVALID)
        elif op == ops.INVOKEINIT:
            old = popped[0]
            if old.tag == '.new':
                new = _indexToCFMInfo(self.cpool, self.instruction[1], 'Method')
            else: # .init
                new = T_OBJECT(self.class_.name)
                self.isThisCtor = True
            state.replace(old, new)

        # Make sure that push happens after local replacement in case of new/invokeinit
        if self.stack_code is not None:
            state.push(popped[i] for i in self.stack_code)
        elif op == ops.ARRLOAD_OBJ:
            # temporary hack
            result = self._removeInterface(decrementDim(popped[0]))
            state.push([result])

        elif op == ops.NEWARRAY or op == ops.ANEWARRAY:
            arrt = self.stack_push[0]
            size = popped[0].const
            if size is not None:
                arrt = exactArrayFrom(arrt, size)
            state.push([arrt])
        else:
            state.push(self.stack_push)

        if self.op in (ops.RET, ops.JSR):
            state.invalidateNews()
            self.out_state = state # store for later convienence

        assert all(isinstance(vt, fullinfo_t) for vt in state.stack)
        assert all(isinstance(vt, fullinfo_t) for vt in state.locals)
        return state

    def _mergeSingleSuccessor(self, other, newstate, iNodes, isException):
        if self.op == ops.RET and not isException:
            # Get the instruction before other
            off_i = self.offsetToIndex[other.key]
            jsrnode = iNodes[self.indexToOffset[off_i-1]]
            jsrnode.returnedFrom = self.key

            if jsrnode.visited: # if not, skip for later
                newstate = newstate.copy()
                newstate.returnTo(jsrnode.instruction[1], jsrnode.state)
            else:
                return

        if not other.visited:
            other.state = newstate.copy()
            other.visited = other.changed = True
        else:
            changed = other.state.merge(newstate, self.env)
            other.changed = other.changed or changed

    def update(self, iNodes, exceptions):
        assert self.visited
        self.changed = False
        newstate = self._getNewState(iNodes)

        successors = self.successors
        if self.op == ops.JSR and self.returnedFrom is not None:
            iNodes[self.returnedFrom].changed = True

        if successors is None:
            assert self.op == ops.RET
            called = self.state.local(self.instruction[1]).extra
            temp = [n.next_instruction for n in iNodes.values() if (n.op == ops.JSR and n.instruction[1] == called)]
            successors = self.successors = tuple(temp)
            self.jsrTarget = called # store for later use in ssa creation

        # Merge into exception handlers first
        for (start, end, handler, except_info) in exceptions:
            if start <= self.key < end:
                self._mergeSingleSuccessor(handler, self.state.withExcept(except_info), iNodes, True)
                if self.op == ops.INVOKEINIT: # two cases since the ctor may suceed or fail before throwing
                    self._mergeSingleSuccessor(handler, newstate.withExcept(except_info), iNodes, True)

        # Now regular successors
        for k in self.successors:
            self._mergeSingleSuccessor(iNodes[k], newstate, iNodes, False)

    def __str__(self):   # pragma: no cover
        lines = ['{}: {}'.format(self.key, bytecode.printInstruction(self.instruction))]
        if self.visited:
            lines.append('Stack: ' + ', '.join(map(str, self.state.stack)))
            lines.append('Locals: ' + ', '.join(map(str, self.state.locals)))
            if self.state.masks:
                lines.append('Masks:')
                lines += ['\t{}: {}'.format(entry, sorted(cset)) for entry, cset in self.state.masks]
        else:
            lines.append('\tunvisited')
        return '\n'.join(lines) + '\n'

def verifyBytecode(code):
    method, class_ = code.method, code.class_
    args, rval = parseUnboundMethodDescriptor(method.descriptor, class_.name, method.static)
    env = class_.env

    # Object has no superclass to construct, so it doesn't get an uninit this
    if method.isConstructor and class_.name != 'java/lang/Object':
        assert args[0] == T_OBJECT(class_.name)
        args[0] = T_UNINIT_THIS
    assert len(args) <= 255 and len(args) <= code.locals

    offsets = sorted(code.bytecode.keys())
    offset_rmap = {v:i for i,v in enumerate(offsets)}
    offsets.append(None) # Sentinel for end of code
    iNodes = [InstructionNode(code, offset_rmap, offsets, key) for key in offsets[:-1]]
    iNodeLookup = {n.key:n for n in iNodes}

    keys = frozenset(iNodeLookup)
    for raw in code.except_raw:
        if not ((0 <= raw.start < raw.end) and (raw.start in keys) and
            (raw.handler in keys) and (raw.end in keys or raw.end == code.codelen)):

            keylist = sorted(keys) + [code.codelen]
            msg = "Illegal exception handler: {}\nValid offsets are: {}".format(raw, ', '.join(map(str, keylist)))
            raise error_types.VerificationError(msg)

    def makeException(rawdata):
        if rawdata.type_ind:
            typen = class_.cpool.getArgsCheck('Class', rawdata.type_ind)
        else:
            typen = 'java/lang/Throwable'
        return (rawdata.start, rawdata.end, iNodeLookup[rawdata.handler], T_OBJECT(typen))
    exceptions = map(makeException, code.except_raw)

    start = iNodes[0]
    start.state = stateFromInitialArgs(args)
    start.visited, start.changed = True, True

    done = False
    while not done:
        done = True
        for node in iNodes:
            if node.changed:
                node.update(iNodeLookup, exceptions)
                done = False
    return iNodes



from ..functionbase import SSAFunctionBase
from ..ssa_types import SSA_OBJECT

class BaseOp(SSAFunctionBase):
    has_side_effects = False

    def __init__(self, parent, arguments, makeException=False):
        super(BaseOp, self).__init__(parent, arguments)

        self.rval = None
        self.outException = None

        if makeException:
            self.outException = parent.makeVariable(SSA_OBJECT, origin=self)

    def getOutputs(self):
        return self.rval, self.outException

    def removeOutput(self, var):
        outs = self.rval, self.outException
        assert var is not None and var in outs
        self.rval, self.outException = [(x if x != var else None) for x in outs]

    def replaceOutVars(self, vardict):
        self.rval, self.outException = map(vardict.get, (self.rval, self.outException))

    # Given input constraints, return constraints on outputs. Output is (rval, exception)
    # With None returned for unused or impossible values. This should only be defined if it is
    # actually implemented.
    # def propagateConstraints(self, *cons):



from ...verifier.descriptors import parseFieldDescriptor

from .. import constraints, excepttypes, objtypes
from ..constraints import IntConstraint, ObjectConstraint, returnOrThrow, throw
from ..ssa_types import SSA_INT, SSA_OBJECT, verifierToSSAType

from .base import BaseOp

# Empirically, Hotspot does enfore size restrictions on short fields
# Except that bool is still a byte
_short_constraints = {
        objtypes.ByteTT: IntConstraint.range(32, -128, 127),
        objtypes.CharTT: IntConstraint.range(32, 0, 65535),
        objtypes.ShortTT: IntConstraint.range(32, -32768, 32767),
        objtypes.IntTT: IntConstraint.bot(32)
    }
_short_constraints[objtypes.BoolTT] = _short_constraints[objtypes.ByteTT]
# Assume no linkage errors occur, so only exception that can be thrown is NPE
class FieldAccess(BaseOp):
    has_side_effects = True

    def __init__(self, parent, instr, info, args):
        super(FieldAccess, self).__init__(parent, args, makeException=('field' in instr[0]))

        self.instruction = instr
        self.target, self.name, self.desc = info

        dtype = None
        if 'get' in instr[0]:
            vtypes = parseFieldDescriptor(self.desc)
            stype = verifierToSSAType(vtypes[0])
            dtype = objtypes.verifierToSynthetic(vtypes[0]) # todo, find way to merge this with Invoke code?
            cat = len(vtypes)

            self.rval = parent.makeVariable(stype, origin=self)
            self.returned = [self.rval] + [None]*(cat-1)
        else:
            self.returned = []

        # just use a fixed constraint until we can do interprocedural analysis
        # output order is rval, exception, defined by BaseOp.getOutputs
        env = parent.env
        self.eout = ObjectConstraint.fromTops(env, [excepttypes.NullPtr], [], nonnull=True)
        if self.rval is not None:
            if self.rval.type == SSA_OBJECT:
                supers, exact = objtypes.declTypeToActual(env, dtype)
                self.rout = ObjectConstraint.fromTops(env, supers, exact)
            elif self.rval.type == SSA_INT:
                self.rout = _short_constraints[dtype]
            else:
                self.rout = constraints.fromVariable(env, self.rval)
        else:
            self.rout = None

    def propagateConstraints(self, *incons):
        eout = None # no NPE
        if 'field' in self.instruction[0] and incons[0].null:
            eout = self.eout
            if incons[0].isConstNull():
                return throw(eout)

        return returnOrThrow(self.rout, eout)



import collections
import itertools

class NameGen(object):
    def __init__(self, reserved=frozenset()):
        self.counters = collections.defaultdict(itertools.count)
        self.names = set(reserved)

    def getPrefix(self, prefix, sep=''):
        newname = prefix
        while newname in self.names:
            newname = prefix + sep + str(next(self.counters[prefix]))        
        self.names.add(newname)
        return newname

def LabelGen(prefix='label'):
    for i in itertools.count():
        yield prefix + str(i)



import itertools

def tarjanSCC(roots, getChildren):
    """Return a list of strongly connected components in a graph. If getParents is passed instead of getChildren, the result will be topologically sorted.

    roots - list of root nodes to search from
    getChildren - function which returns children of a given node
    """

    sccs = []
    indexCounter = itertools.count()
    index = {}
    lowlink = {}
    removed = set()
    subtree = []

    # Use iterative version to avoid stack limits for large datasets
    stack = [(node, 0) for node in roots]
    while stack:
        current, state = stack.pop()
        if state == 0: # before recursing
            if current not in index: # if it's in index, it was already visited (possibly earlier on the current search stack)
                lowlink[current] = index[current] = next(indexCounter)
                subtree.append(current)

                stack.append((current, 1))
                stack.extend((child, 0) for child in getChildren(current) if child not in removed)
        else: # after recursing
            children = [child for child in getChildren(current) if child not in removed]
            for child in children:
                if index[child] <= index[current]: # backedge (or selfedge)
                    lowlink[current] = min(lowlink[current], index[child])
                else:
                    lowlink[current] = min(lowlink[current], lowlink[child])
                assert lowlink[current] <= index[current]

            if index[current] == lowlink[current]:
                scc = []
                while not scc or scc[-1] != current:
                    scc.append(subtree.pop())

                sccs.append(tuple(scc))
                removed.update(scc)
    return sccs

def topologicalSort(roots, getParents):
    """Return a topological sorting of nodes in a graph.

    roots - list of root nodes to search from
    getParents - function which returns the parents of a given node
    """

    results = []
    visited = set()

    # Use iterative version to avoid stack limits for large datasets
    stack = [(node,0) for node in roots]
    while stack:
        current, state = stack.pop()
        if state == 0: # before recursing
            if current not in visited:
                visited.add(current)
                stack.append((current,1))
                stack.extend((parent,0) for parent in getParents(current))
        else: # after recursing
            assert current in visited
            results.append(current)
    return results



from .. import ssa_types
from ..constraints import IntConstraint, ObjectConstraint

from .base import BaseJump
from .goto import Goto

class If(BaseJump):
    opposites = {'eq':'ne', 'ne':'eq', 'lt':'ge', 'ge':'lt', 'gt':'le', 'le':'gt'}

    def __init__(self, parent, cmp, successors, arguments):
        super(If, self).__init__(parent, arguments)
        assert cmp in ('eq','ne','lt','ge','gt','le')
        self.cmp = cmp
        self.successors = successors
        self.isObj = (arguments[0].type == ssa_types.SSA_OBJECT)
        assert None not in successors

    def replaceBlocks(self, blockDict):
        self.successors = [blockDict.get(key,key) for key in self.successors]

    def getNormalSuccessors(self):
        return self.successors

    def reduceSuccessors(self, pairsToRemove):
        temp = set(self.successors)
        for (child, t) in pairsToRemove:
            temp.remove(child)

        if len(temp) == 0:
            return None
        elif len(temp) == 1:
            return Goto(self.parent, temp.pop())
        return self

    ###############################################################################
    def constrainJumps(self, x, y):
        impossible = []
        for child in self.successors:
            func = self.getSuccessorConstraints((child,False))

            results = func(x,y)
            if None in results:
                assert results == (None,None)
                impossible.append((child,False))
        return self.reduceSuccessors(impossible)

    def getSuccessorConstraints(self, xxx_todo_changeme):
        (block, t) = xxx_todo_changeme
        assert t is False
        cmp_t = If.opposites[self.cmp] if block == self.successors[0] else self.cmp

        if self.isObj:
            def propagateConstraints_obj(x, y):
                if x is None or y is None:
                    return None, None
                if cmp_t == 'eq':
                    z = x.join(y)
                    return z,z
                else:
                    x2, y2 = x, y
                    if x.isConstNull():
                        yt = y.types
                        y2 = ObjectConstraint.fromTops(yt.env, yt.supers, yt.exact, nonnull=True)
                    if y.isConstNull():
                        xt = x.types
                        x2 = ObjectConstraint.fromTops(xt.env, xt.supers, xt.exact, nonnull=True)
                    return x2, y2
            return propagateConstraints_obj
        else:
            def propagateConstraints_int(x, y):
                if x is None or y is None:
                    return None, None
                x1, x2, y1, y2 = x.min, x.max, y.min, y.max
                if cmp_t == 'ge' or cmp_t == 'gt':
                    x1, x2, y1, y2 = y1, y2, x1, x2

                # treat greater like less than swap before and afterwards
                if cmp_t == 'lt' or cmp_t == 'gt':
                    x2 = min(x2, y2-1)
                    y1 = max(x1+1, y1)
                elif cmp_t == 'le' or cmp_t == 'ge':
                    x2 = min(x2, y2)
                    y1 = max(x1, y1)
                elif cmp_t == 'eq':
                    x1 = y1 = max(x1, y1)
                    x2 = y2 = min(x2, y2)
                elif cmp_t == 'ne':
                    if x1 == x2 == y1 == y2:
                        return None, None
                    if x1 == x2:
                        y1 = y1 if y1 != x1 else y1+1
                        y2 = y2 if y2 != x2 else y2-1
                    if y1 == y2:
                        x1 = x1 if x1 != y1 else x1+1
                        x2 = x2 if x2 != y2 else x2-1

                if cmp_t == 'ge' or cmp_t == 'gt':
                    x1, x2, y1, y2 = y1, y2, x1, x2
                con1 = IntConstraint.range(x.width, x1, x2) if x1 <= x2 else None
                con2 = IntConstraint.range(y.width, y1, y2) if y1 <= y2 else None
                return con1, con2
            return propagateConstraints_int



import itertools

def update(self, items):
    self.entryBlock = items[0].entryBlock
    self.nodes = frozenset.union(*(i.nodes for i in items))
    temp = set(self.nodes)
    siter = itertools.chain.from_iterable(i.successors for i in items)
    self.successors = [n for n in siter if not n in temp and not temp.add(n)]

class SEBlockItem(object):
    def __init__(self, node):
        self.successors = node.norm_suc_nl # don't include backedges or exceptional edges
        self.node = node
        self.nodes = frozenset([node])
        self.entryBlock = node

    def getScopes(self): return ()

class SEScope(object):
    def __init__(self, items):
        self.items = items
        update(self, items)

    def getScopes(self): return ()

class SEWhile(object):
    def __init__(self, scope):
        self.body = scope
        update(self, [scope])

    def getScopes(self): return self.body,

class SETry(object):
    def __init__(self, tryscope, catchscope, toptts, catchvar):
        self.scopes = tryscope, catchscope
        self.toptts = toptts
        self.catchvar = catchvar # none if ignored
        update(self, self.scopes)

    def getScopes(self): return self.scopes

class SEIf(object):
    def __init__(self, head, newscopes):
        assert len(newscopes) == 2
        self.scopes = newscopes
        self.head = head
        update(self, [head] + newscopes)

    def getScopes(self): return self.scopes

class SESwitch(object):
    def __init__(self, head, newscopes):
        self.scopes = newscopes
        self.head = head
        self.ordered = newscopes
        update(self, [head] + newscopes)

        jump = head.node.block.jump
        keysets = {head.node.blockdict[b.key,False]:jump.reverse.get(b) for b in jump.getNormalSuccessors()}
        assert keysets.values().count(None) == 1
        self.ordered_keysets = [keysets[item.entryBlock] for item in newscopes]

    def getScopes(self): return self.scopes



from collections import namedtuple as nt

# Define types for Inference
# Extra stores address for .new and .address and class name for object types
# Const stores value for int constants and length for exact arrays. This isn't needed for normal verification but is
# useful for optimizing the code later.
fullinfo_t = nt('fullinfo_t', ['tag','dim','extra','const'])

valid_tags = ['.'+_x for _x in 'int float double long obj new init address byte short char boolean'.split()]
valid_tags = frozenset([None] + valid_tags)

def _makeinfo(tag, dim=0, extra=None, const=None):
    assert tag in valid_tags
    return fullinfo_t(tag, dim, extra, const)

T_INVALID = _makeinfo(None)
T_INT = _makeinfo('.int')
T_FLOAT = _makeinfo('.float')
T_DOUBLE = _makeinfo('.double')
T_LONG = _makeinfo('.long')

T_NULL = _makeinfo('.obj')
T_UNINIT_THIS = _makeinfo('.init')

T_BYTE = _makeinfo('.byte')
T_SHORT = _makeinfo('.short')
T_CHAR = _makeinfo('.char')
T_BOOL = _makeinfo('.boolean') # Hotspot doesn't have a bool type, but we can use this elsewhere

# types with arguments
def T_ADDRESS(entry):
    return _makeinfo('.address', extra=entry)

def T_OBJECT(name):
    return _makeinfo('.obj', extra=name)

def T_ARRAY(baset, newDimensions=1):
    assert 0 <= baset.dim <= 255-newDimensions
    return _makeinfo(baset.tag, baset.dim+newDimensions, baset.extra)

def T_UNINIT_OBJECT(origin):
    return _makeinfo('.new', extra=origin)

def T_INT_CONST(val):
    assert -0x80000000 <= val < 0x80000000
    return _makeinfo(T_INT.tag, const=val)

OBJECT_INFO = T_OBJECT('java/lang/Object')
CLONE_INFO = T_OBJECT('java/lang/Cloneable')
SERIAL_INFO = T_OBJECT('java/io/Serializable')
THROWABLE_INFO = T_OBJECT('java/lang/Throwable')

def objOrArray(fi): # False on uninitialized
    return fi.tag == '.obj' or fi.dim > 0

def unSynthesizeType(t):
    if t in (T_BOOL, T_BYTE, T_CHAR, T_SHORT):
        return T_INT
    return t

def decrementDim(fi):
    if fi == T_NULL:
        return T_NULL
    assert fi.dim

    tag = unSynthesizeType(fi).tag if fi.dim <= 1 else fi.tag
    return _makeinfo(tag, fi.dim-1, fi.extra)

def exactArrayFrom(fi, size):
    assert fi.dim > 0
    if size >= 0:
        return _makeinfo(fi.tag, fi.dim, fi.extra, size)
    return fi

def withNoDimension(fi):
    return _makeinfo(fi.tag, 0, fi.extra)

def withNoConst(fi):
    if fi.const is None:
        return fi
    return _makeinfo(fi.tag, fi.dim, fi.extra)

def _decToObjArray(fi):
    return fi if fi.tag == '.obj' else T_ARRAY(OBJECT_INFO, fi.dim-1)

def mergeTypes(env, t1, t2):
    if t1 == t2:
        return t1

    t1 = withNoConst(t1)
    t2 = withNoConst(t2)
    if t1 == t2:
        return t1

    # non objects must match exactly
    if not objOrArray(t1) or not objOrArray(t2):
        return T_INVALID

    if t1 == T_NULL:
        return t2
    elif t2 == T_NULL:
        return t1

    if t1 == OBJECT_INFO or t2 == OBJECT_INFO:
        return OBJECT_INFO

    if t1.dim or t2.dim:
        for x in (t1,t2):
            if x in (CLONE_INFO,SERIAL_INFO):
                return x
        t1 = _decToObjArray(t1)
        t2 = _decToObjArray(t2)

        if t1.dim > t2.dim:
            t1, t2 = t2, t1

        if t1.dim == t2.dim:
            res = mergeTypes(env, withNoDimension(t1), withNoDimension(t2))
            return res if res == T_INVALID else _makeinfo('.obj', t1.dim, res.extra)
        else: # t1.dim < t2.dim
            return t1 if withNoDimension(t1) in (CLONE_INFO, SERIAL_INFO) else T_ARRAY(OBJECT_INFO, t1.dim)
    else: # neither is array
        if env.isInterface(t2.extra, forceCheck=True):
            return OBJECT_INFO
        return T_OBJECT(env.commonSuperclass(t1.extra, t2.extra))


# Make verifier types printable for easy debugging
def vt_toStr(self):   # pragma: no cover
    if self == T_INVALID:
        return '.none'
    elif self == T_NULL:
        return '.null'
    if self.tag == '.obj':
        base = self.extra
    elif self.extra is not None:
        base = '{}<{}>'.format(self.tag, self.extra)
    else:
        base = self.tag
    return base + '[]'*self.dim
fullinfo_t.__str__ = fullinfo_t.__repr__ = vt_toStr



from .base import BaseJump

from .onexception import OnException
from .goto import Goto
from .ifcmp import If
from .exit import Return, Rethrow
from .switch import Switch

from . import placeholder
OnAbscond = Ret = placeholder.Placeholder



from ..ssa import objtypes

from . import ast

# A simple throws declaration inferrer that only considers throw statements within the method
# this is mostly just useful to make sure the ExceptionHandlers test compiles

def _visit_statement(env, stmt):
    if isinstance(stmt, ast.ThrowStatement):
        return stmt.expr.dtype

    result = objtypes.NullTT
    if isinstance(stmt, ast.TryStatement):
        caught_types = []
        for catch, b in stmt.pairs:
            caught_types.extend(objtypes.className(tn.tt) for tn in catch.typename.tnames)
        if objtypes.ThrowableTT not in caught_types:
            temp = _visit_statement(env, stmt.tryb)
            if temp != objtypes.NullTT:
                assert objtypes.dim(temp) == 0
                name = objtypes.className(temp)
                if not any(env.isSubclass(name, caught) for caught in caught_types):
                    result = temp

        statements = zip(*stmt.pairs)[1]
    elif isinstance(stmt, ast.StatementBlock):
        statements = stmt.statements
    else:
        statements = stmt.getScopes()

    for sub in statements:
        if result == objtypes.ThrowableTT:
            break
        result = objtypes.commonSupertype(env, [result, _visit_statement(env, sub)])

    if result != objtypes.NullTT:
        if env.isSubclass(objtypes.className(result), 'java/lang/RuntimeException'):
            return objtypes.NullTT
    return result

def addSingle(env, meth_asts):
    for meth in meth_asts:
        if not meth.body:
            continue
        tt = _visit_statement(env, meth.body)
        assert objtypes.commonSupertype(env, [tt, objtypes.ThrowableTT]) == objtypes.ThrowableTT
        if tt != objtypes.NullTT:
            meth.throws = ast.TypeName(tt)



from .base import BaseOp

class Phi(object):
    __slots__ = 'block dict rval'.split()
    has_side_effects = False

    def __init__(self, block, rval):
        self.block = block # used in constraint propagation
        self.dict = {}
        self.rval = rval
        assert rval is not None and rval.origin is None
        rval.origin = self

    def add(self, key, val):
        assert key not in self.dict
        assert val.type == self.rval.type
        assert val is not None
        self.dict[key] = val

    @property
    def params(self): return [self.dict[k] for k in self.block.predecessors]

    def get(self, key): return self.dict[key]
    def delete(self, key): del self.dict[key]

    # Copy these over from BaseOp so we don't need to inherit
    def replaceVars(self, rdict):
        for k in self.dict:
            self.dict[k] = rdict.get(self.dict[k], self.dict[k])

    def getOutputs(self):
        return self.rval, None, None

    def removeOutput(self, var):
        assert var == self.rval
        self.rval = None

# An extended basic block can contain multiple throwing instructions
# but the OnException jump expects a single param. The solution is
# to create a dummy op that behaves like a phi function, selecting
# among the possible thrown exceptions in the block. This is always
# the last op in block.lines when there are exceptions.
# As this is a phi, params can be variable length
class ExceptionPhi(BaseOp):
    def __init__(self, parent, params):
        BaseOp.__init__(self, parent, params, makeException=True)



from __future__ import print_function

import collections
import errno
from functools import partial
import hashlib
import os
import os.path
import platform
import zipfile

# Various utility functions for the top level scripts (decompile.py, assemble.py, disassemble.py)

copyright = '''Krakatau  Copyright (C) 2012-18  Robert Grosse
This program is provided as open source under the GNU General Public License.
See LICENSE.TXT for more details.
'''

_osname = platform.system().lower()
IS_WINDOWS = 'win' in _osname and 'darwin' not in _osname and 'cygwin' not in _osname

def findFiles(target, recursive, prefix):
    if target.endswith('.jar'):
        with zipfile.ZipFile(target, 'r') as archive:
            return [name.encode('utf8') for name in archive.namelist() if name.endswith(prefix)]
    else:
        if recursive:
            assert os.path.isdir(target)
            targets = []

            for root, dirs, files in os.walk(target):
                targets += [os.path.join(root, fname) for fname in files if fname.endswith(prefix)]
            return targets
        else:
            return [target]

def normalizeClassname(name):
    if name.endswith('.class'):
        name = name[:-6]
    # Replacing backslashes is ugly since they can be in valid classnames too, but this seems the best option
    return name.replace('\\','/').replace('.','/')

# Windows stuff
illegal_win_chars = frozenset('<>;:|?*\\/"%')
pref_disp_chars = frozenset('abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ_$0123456789')

# Prevent creating filename parts matching the legacy device filenames. While Krakatau can create these files
# just fine thanks to using \\?\ paths, the resulting files are impossible to open or delete in Windows Explorer
# or with similar tools, so they are a huge pain to deal with. Therefore, we don't generate them at all.
illegal_parts = frozenset(['CON', 'PRN', 'AUX', 'NUL', 'COM1', 'COM2', 'COM3', 'COM4', 'COM5', 'COM6', 'COM7', 'COM8',
    'COM9', 'LPT1', 'LPT2', 'LPT3', 'LPT4', 'LPT5', 'LPT6', 'LPT7', 'LPT8', 'LPT9'])

class PathSanitizer(object):
    def __init__(self, base, suffix):
        self.base = base
        self.suffix = suffix

    def is_part_ok(self, s, parents):
        if not 1 <= len(s) <= self.MAX_PART_LEN:
            return False
        # avoid potential collision with hashed parts
        if len(s) >= 66 and '__' in s:
            return False
        # . cannot appear in a valid class name, but might as well specifically exclude these, just in case
        if s.startswith('.') or '..' in s:
            return False
        return '\x1f' < min(s) <= max(s) < '\x7f'

    def hash(self, s, suffix):
        left = ''.join(c for c in s if c in pref_disp_chars)
        right = '__' + hashlib.sha256(s.encode('utf8')).hexdigest() + suffix
        return left[:self.MAX_PART_LEN - len(right)] + right

    def sanitize(self, path):
        if isinstance(path, bytes):
            path = path.decode()

        oldparts = path.split('/')
        newparts = []
        for i, part in enumerate(oldparts):
            suffix = self.suffix if i + 1 == len(oldparts) else ''
            if self.is_part_ok(part + suffix, newparts):
                newparts.append(part + suffix)
            else:
                newparts.append(self.hash(part, suffix))

        result = self.format_path([self.base] + newparts)
        if len(result) > self.MAX_PATH_LEN:
            result = self.format_path([self.base, self.hash(path, self.suffix)])
        assert result.endswith(self.suffix)
        return result

class LinuxPathSanitizer(PathSanitizer):
    MAX_PART_LEN = 255
    MAX_PATH_LEN = 4095

    def __init__(self, *args):
        PathSanitizer.__init__(self, *args)

    def format_path(self, parts):
        return os.path.join(*parts)

class WindowsPathSanitizer(PathSanitizer):
    MAX_PART_LEN = 255
    MAX_PATH_LEN = 32000 # close enough

    def __init__(self, *args):
        PathSanitizer.__init__(self, *args)
        # keep track of previous paths to detect case-insensitive collisions
        self.prevs = collections.defaultdict(dict)

    def is_part_ok(self, s, parents):
        if not PathSanitizer.is_part_ok(self, s, parents):
            return False
        if s.upper() in illegal_parts:
            return False
        # make sure nothing in the current directory is a case insensitive collision
        if self.prevs[tuple(parents)].setdefault(s.lower(), s) != s:
            return False
        return illegal_win_chars.isdisjoint(s)

    def format_path(self, parts):
        return '\\\\?\\' + '\\'.join(parts)

class DirectoryWriter(object):
    def __init__(self, base_path, suffix):
        if base_path is None:
            base_path = os.getcwd()
        else:
            if not isinstance(base_path, str):
                base_path = base_path.decode('utf8')
            base_path = os.path.abspath(base_path)

        if IS_WINDOWS:
            self.makepath = WindowsPathSanitizer(base_path, suffix).sanitize
        else:
            self.makepath = LinuxPathSanitizer(base_path, suffix).sanitize

    def write(self, cname, data):
        out = self.makepath(cname)
        dirpath = os.path.dirname(out)

        try:
            if dirpath:
                os.makedirs(dirpath)
        except OSError as exc:
            if exc.errno != errno.EEXIST:
                raise

        mode = 'wb' if isinstance(data, bytes) else 'w'
        with open(out, mode) as f:
            f.write(data)
        return out

    def __enter__(self): return self
    def __exit__(self, *args): pass

class JarWriter(object):
    def __init__(self, base_path, suffix):
        self.zip = zipfile.ZipFile(base_path, mode='w')
        self.suffix = suffix

    def write(self, cname, data):
        info = zipfile.ZipInfo(cname + self.suffix, (1980, 1, 1, 0, 0, 0))
        self.zip.writestr(info, data)
        return 'zipfile'

    def __enter__(self): self.zip.__enter__(); return self
    def __exit__(self, *args): self.zip.__exit__(*args)

class MockWriter(object):
    def __init__(self): self.results = []
    def write(self, cname, data): self.results.append((cname, data))
    def __enter__(self): return self
    def __exit__(self, *args): pass

def makeWriter(base_path, suffix):
    if base_path is not None:
        if base_path.endswith('.zip') or base_path.endswith('.jar'):
            return JarWriter(base_path, suffix)
    return DirectoryWriter(base_path, suffix)

###############################################################################
def ignore(*args, **kwargs):
    pass

class Logger(object):
    def __init__(self, level):
        lvl = ['info', 'warning'].index(level)
        self.info = print if lvl <= 0 else ignore
        self.warn = print if lvl <= 1 else ignore



from .pool import Pool, utf
from .writer import Writer, Label


def writeU16Count(data, error, objects, message):
    count = len(objects)
    if count >= 1<<16:
        error('Maximum {} count is {}, found {}.'.format(message, (1<<16)-1, count), objects[-1].tok)
    data.u16(count)

class Code(object):
    def __init__(self, tok, short):
        self.tok = tok
        self.short = short
        self.locals = self.stack = 0

        self.bytecode = Writer()
        self.exceptions = Writer()
        self.exceptcount = 0

        self.stackdata = Writer()
        self.stackcount = 0
        self.stackcountpos = self.stackdata.ph16()
        self.laststackoff = -1 # first frame doesn't subtract 1 from offset

        self.stackmaptable = None
        self.dont_generate_stackmap = False
        self.attributes = []

        self.labels = {}
        self.maxcodelen = (1<<16 if short else 1<<32) - 1

    def labeldef(self, lbl, error):
        if lbl.sym in self.labels:
            error('Duplicate label definition', lbl.tok,
                  'Previous definition here:', self.labels[lbl.sym][0])
        self.labels[lbl.sym] = lbl.tok, self.bytecode.pos

    def catch(self, ref, fromlbl, tolbl, usinglbl):
        self.exceptcount += 1
        self.exceptions.lbl(fromlbl, 0, 'u16')
        self.exceptions.lbl(tolbl, 0, 'u16')
        self.exceptions.lbl(usinglbl, 0, 'u16')
        self.exceptions.ref(ref)

    def assembleNoCP(self, data, error):
        bytecode = self.bytecode

        if self.short:
            data.u8(self.stack), data.u8(self.locals), data.u16(len(bytecode))
        else:
            data.u16(self.stack), data.u16(self.locals), data.u32(len(bytecode))
        data += bytecode

        data.u16(self.exceptcount)
        data += self.exceptions

        if self.stackmaptable is None and self.stackcount > 0 and not self.dont_generate_stackmap:
            # Use arbitrary token in case we need to report errors
            self.stackmaptable = Attribute(self.tok, b'StackMapTable')
            self.attributes.append(self.stackmaptable)

        if self.stackmaptable:
            self.stackdata.setph16(self.stackcountpos, self.stackcount)
            self.stackmaptable.data = self.stackdata

        writeU16Count(data, error, self.attributes, 'attribute')
        for attr in self.attributes:
            attr.assembleNoCP(data, error)
        return data.fillLabels(self.labels, error)

class Attribute(object):
    def __init__(self, tok, name, length=None):
        assert tok
        if isinstance(name, bytes):
            name = utf(tok, name)

        self.tok = tok
        self.name = name
        self.length = length
        self.data = Writer()

    def assembleNoCP(self, data, error):
        length = len(self.data) if self.length is None else self.length
        if length >= 1<<32:
            error('Maximum attribute data length is {} bytes, got {} bytes.'.format((1<<32)-1, length), self.tok)

        data.ref(self.name)
        data.u32(length)
        data += self.data
        return data

class Method(object):
    def __init__(self, tok, access, name, desc):
        self.tok = tok
        self.access = access
        self.name = name
        self.desc = desc
        self.attributes = []

    def assembleNoCP(self, data, error):
        data.u16(self.access)
        data.ref(self.name)
        data.ref(self.desc)

        writeU16Count(data, error, self.attributes, 'attribute')
        for attr in self.attributes:
            attr.assembleNoCP(data, error)
        return data

class Field(object):
    def __init__(self, tok, access, name, desc):
        self.tok = tok
        self.access = access
        self.name = name
        self.desc = desc
        self.attributes = []

    def assembleNoCP(self, data, error):
        data.u16(self.access)
        data.ref(self.name)
        data.ref(self.desc)

        writeU16Count(data, error, self.attributes, 'attribute')
        for attr in self.attributes:
            attr.assembleNoCP(data, error)
        return data

class Class(object):
    def __init__(self):
        self.version = 49, 0
        self.access = self.this = self.super = None

        self.interfaces = []
        self.fields = []
        self.methods = []
        self.attributes = []

        self.useshortcodeattrs = False
        self.bootstrapmethods = None
        self.pool = Pool()

    def _getName(self):
        cpool = self.pool.cp
        clsind = self.this.resolved_index
        if not cpool.slots.get(clsind):
            return None

        if cpool.slots[clsind].type != 'Class':
            return None

        utfind = cpool.slots[clsind].refs[0].resolved_index
        if utfind not in cpool.slots:
            return None

        return cpool.slots[utfind].data

    def _assembleNoCP(self, error):
        beforepool = Writer()
        afterpool = Writer()

        beforepool.u32(0xCAFEBABE)
        beforepool.u16(self.version[1])
        beforepool.u16(self.version[0])

        afterpool.u16(self.access)
        afterpool.ref(self.this)
        afterpool.ref(self.super)
        writeU16Count(afterpool, error, self.interfaces, 'interface')
        for i in self.interfaces:
            afterpool.ref(i)

        writeU16Count(afterpool, error, self.fields, 'field')
        for field in self.fields:
            field.assembleNoCP(afterpool, error)

        writeU16Count(afterpool, error, self.methods, 'method')
        for method in self.methods:
            method.assembleNoCP(afterpool, error)

        attrcountpos = afterpool.ph16()
        afterbs = Writer()

        data = afterpool
        for attr in self.attributes:
            if attr is self.bootstrapmethods:
                # skip writing this attr for now and switch to after bs stream
                data = afterbs
            else:
                attr.assembleNoCP(data, error)

        return beforepool, afterpool, afterbs, attrcountpos

    def assemble(self, error):
        beforepool, afterpool, afterbs, attrcountpos = self._assembleNoCP(error)

        self.pool.cp.freezedefs(self.pool, error)
        self.pool.bs.freezedefs(self.pool, error)

        # afterpool is the only part that can contain ldcs
        assert not beforepool.refu8phs
        assert not afterbs.refu8phs
        for _, ref in afterpool.refu8phs:
            ind = ref.resolve(self.pool, error)
            if ind >= 256:
                error("Ldc references too many distinct constants in this class. If you don't want to see this message again, use ldc_w instead of ldc everywhere.", ref.tok)

        beforepool.fillRefs(self.pool, error)
        afterpool.fillRefs(self.pool, error)
        afterbs.fillRefs(self.pool, error)

        # Figure out if we need to add an implicit BootstrapMethods attribute
        self.pool.resolveIDBSRefs(error)
        if self.bootstrapmethods is None and self.pool.bs.slots:
            assert len(afterbs) == 0
            # Use arbitrary token in case we need to report errors
            self.bootstrapmethods = Attribute(self.this.tok, b'BootstrapMethods')
            self.attributes.append(self.bootstrapmethods)
        if self.bootstrapmethods is not None:
            self.bootstrapmethods.name.resolve(self.pool, error)
            assert len(self.bootstrapmethods.data) == 0

        if len(self.attributes) >= 1<<16:
            error('Maximum class attribute count is 65535, found {}.'.format(count), self.attributes[-1].tok)
        afterpool.setph16(attrcountpos, len(self.attributes))


        cpdata, bsmdata = self.pool.write(error)
        assert len(bsmdata) < (1 << 32)

        data = beforepool
        data += cpdata
        data += afterpool

        if self.bootstrapmethods is not None:
            self.bootstrapmethods.data = bsmdata
            self.bootstrapmethods.assembleNoCP(data, error)
            data.fillRefs(self.pool, error)
            data += afterbs
        else:
            assert len(afterbs) == 0

        name = self._getName()
        if name is None:
            error('Invalid reference for class name.', self.this.tok)
        return name, data.toBytes()



import collections, itertools

from ... import floatutil
from .. import objtypes
from .int_c import IntConstraint
from .float_c import FloatConstraint
from .obj_c import ObjectConstraint

from ..ssa_types import SSA_INT, SSA_LONG, SSA_FLOAT, SSA_DOUBLE, SSA_OBJECT

# joins become more precise (intersection), meets become more general (union)
# Join currently supports joining a max of two constraints
# Meet assumes all inputs are not None
def join(*cons):
    if None in cons:
        return None
    return cons[0].join(*cons[1:])

def meet(*cons):
    if not cons:
        return None
    return cons[0].meet(*cons[1:])

def fromConstant(env, var):
    ssa_type = var.type
    cval = var.const

    if ssa_type[0] == SSA_INT[0]:
        return IntConstraint.const(ssa_type[1], cval)
    elif ssa_type[0] == SSA_FLOAT[0]:
        xt = floatutil.fromRawFloat(ssa_type[1], cval)
        return FloatConstraint.const(ssa_type[1], xt)
    elif ssa_type[0] == SSA_OBJECT[0]:
        if var.decltype == objtypes.NullTT:
            return ObjectConstraint.constNull(env)
        return ObjectConstraint.fromTops(env, *objtypes.declTypeToActual(env, var.decltype))

_bots = {
    SSA_INT: IntConstraint.bot(SSA_INT[1]),
    SSA_LONG: IntConstraint.bot(SSA_LONG[1]),
    SSA_FLOAT: FloatConstraint.bot(SSA_FLOAT[1]),
    SSA_DOUBLE: FloatConstraint.bot(SSA_DOUBLE[1]),
}

def fromVariable(env, var):
    if var.const is not None:
        return fromConstant(env, var)
    ssa_type = var.type

    try:
        return _bots[ssa_type]
    except KeyError:
        assert ssa_type == SSA_OBJECT
        isnew = var.uninit_orig_num is not None
        if var.decltype is not None:
            if var.decltype == objtypes.NullTT:
                return ObjectConstraint.constNull(env)
            return ObjectConstraint.fromTops(env, *objtypes.declTypeToActual(env, var.decltype), nonnull=isnew)
        else:
            return ObjectConstraint.fromTops(env, [objtypes.ObjectTT], [], nonnull=isnew)

OpReturnInfo = collections.namedtuple('OpReturnInfo', ['rval', 'eval', 'must_throw'])
def returnOrThrow(rval, eval): return OpReturnInfo(rval, eval, False)
def maybeThrow(eval): return OpReturnInfo(None, eval, False)
def throw(eval): return OpReturnInfo(None, eval, True)
def return_(rval): return OpReturnInfo(rval, None, False)



from collections import defaultdict as ddict
import itertools

def unique(seq): return len(set(seq)) == len(seq)

# This module provides a view of the ssa graph that can be modified without
# touching the underlying graph. This proxy is tailored towards the need of
# cfg structuring, so it allows easy duplication and indirection of nodes,
# but assumes that the underlying variables and statements are immutable

class BlockProxy(object):
    def __init__(self, key, counter, block=None):
        self.bkey = key
        self.num = next(counter)
        self.counter = counter
        self.block = block

        self.predecessors = []
        self.successors = []
        self.outvars = {}
        self.eassigns = {} # exception edge assignments, used after try constraint creation
        self._key = self.bkey, self.num

        # to be assigned later
        self.invars = self.blockdict = None
        # assigned by structuring.py calcNoLoopNeighbors
        self.successors_nl = self.predecessors_nl = self.norm_suc_nl = None

    def replaceSuccessors(self, rmap):
        update = lambda k:rmap.get(k,k)

        self.successors = map(update, self.successors)
        self.outvars = {update(k):v for k,v in self.outvars.items()}
        if self.block is not None:
            d1 = self.blockdict
            self.blockdict = {(b.key,t):update(d1[b.key,t]) for (b,t) in self.block.jump.getSuccessorPairs()}

    def newIndirect(self): # for use during graph creation
        new = BlockProxy(self.bkey, self.counter)
        new.invars = self.invars
        new.outvars = {self:new.invars}
        new.blockdict = None
        new.successors = [self]
        self.predecessors.append(new)
        return new

    def newDuplicate(self): # for use by structuring.structure return inlining
        new = BlockProxy(self.bkey, self.counter, self.block)
        new.invars = self.invars
        new.outvars = self.outvars.copy()
        new.blockdict = self.blockdict
        new.successors = self.successors[:]
        return new

    def indirectEdges(self, edges):
        # Should only be called once graph is completely set up. newIndirect is used during graph creation
        new = self.newIndirect()
        for parent in edges:
            self.predecessors.remove(parent)
            new.predecessors.append(parent)
            parent.replaceSuccessors({self:new})
        return new

    def normalSuccessors(self): # only works once try constraints have been created
        return [x for x in self.successors if x in self.outvars]

    def __str__(self):   # pragma: no cover
        fmt = 'PB {}x{}' if self.num else 'PB {0}'
        return fmt.format(self.bkey, self.num)
    __repr__ = __str__


def createGraphProxy(ssagraph):
    assert(not ssagraph.procs) # should have already been inlined

    nodes = [BlockProxy(b.key, itertools.count(), block=b) for b in ssagraph.blocks]
    allnodes = nodes[:] # will also contain indirected nodes

    entryNode = None
    intypes = ddict(set)
    for n in nodes:
        invars = [phi.rval for phi in n.block.phis]
        for b, t in n.block.jump.getSuccessorPairs():
            intypes[b.key].add(t)

        if n.bkey == ssagraph.entryKey:
            assert(not entryNode and not invars) # shouldn't have more than one entryBlock and entryBlock shouldn't have phis
            entryNode = n
            invars = ssagraph.inputArgs # store them in the node so we don't have to keep track seperately
            invars = [x for x in invars if x is not None] # will have None placeholders for Long and Double arguments
        n.invars = invars

    lookup = {}
    for n in nodes:
        assert len(intypes[n.bkey]) != 2 # should have been handled by graph.splitDualInedges()

        if False in intypes[n.bkey]:
            lookup[n.bkey, False] = n
        if True in intypes[n.bkey]:
            lookup[n.bkey, True] = n
    assert unique(lookup.values())

    for n in nodes:
        n.blockdict = lookup
        block = n.block
        for (block2, t) in block.jump.getSuccessorPairs():
            out = [phi.get((block, t)) for phi in block2.phis]

            n2 = lookup[block2.key, t]
            n.outvars[n2] = out
            n.successors.append(n2)
            n2.predecessors.append(n)

    # sanity check
    for n in allnodes:
        assert (n.block is not None) == (n.num == 0)
        assert (n is entryNode) == (len(n.predecessors) == 0)
        assert unique(n.predecessors)
        assert unique(n.successors)
        for pn in n.predecessors:
            assert n in pn.successors
        assert set(n.outvars) == set(n.successors)
        for sn in n.successors:
            assert n in sn.predecessors
            assert len(n.outvars[sn]) == len(sn.invars)

    return entryNode, allnodes



import collections
import copy
import functools
import itertools

from .. import graph_util
from ..verifier.descriptors import parseUnboundMethodDescriptor

from . import blockmaker, constraints, objtypes, ssa_jumps, ssa_ops, subproc
from .ssa_types import BasicBlock, SSA_OBJECT, verifierToSSAType

class SSA_Variable(object):
    __slots__ = 'type','origin','name','const','decltype','uninit_orig_num'

    def __init__(self, type_, origin=None, name=""):
        self.type = type_       # SSA_INT, SSA_OBJECT, etc.
        self.origin = origin
        self.name = name
        self.const = None
        self.decltype = None # for objects, the inferred type from the verifier if any
        self.uninit_orig_num = None # if uninitialized, the bytecode offset of the new instr

    # for debugging
    def __str__(self):   # pragma: no cover
        return self.name if self.name else super(SSA_Variable, self).__str__()

    def __repr__(self):   # pragma: no cover
        name =  self.name if self.name else "@" + hex(id(self))
        return "Var {}".format(name)

# This class is the main IR for bytecode level methods. It consists of a control
# flow graph (CFG) in static single assignment form (SSA). Each node in the
# graph is a BasicBlock. This consists of a list of phi statements representing
# inputs, a list of operations, and a jump statement. Exceptions are represented
# explicitly in the graph with the OnException jump. Each block also keeps track
# of the unary constraints on the variables in that block.

# Handling of subprocedures is rather annoying. Each complete subproc has an associated
# ProcInfo while jsrs and rets are represented by ProcCallOp and DummyRet respectively.
# The jsrblock has the target and fallthrough as successors, while the fallthrough has
# the jsrblock as predecessor, but not the retblock. Control flow paths where the proc
# never returns are represented by ordinary jumps from blocks in the procedure to outside
# Successful completion of the proc is represented by the fallthrough edge. The fallthrough
# block gets its variables from the jsrblock, including skip vars which don't depend on the
# proc, and variables from jsr.output which represent what would have been returned from ret
# Every proc has a reachable retblock. Jsrs with no associated ret are simply turned
# into gotos during the initial basic block creation.

class SSA_Graph(object):
    entryKey = blockmaker.ENTRY_KEY

    def __init__(self, code):
        self.code = code
        self.class_ = code.class_
        self.env = self.class_.env

        self.inputArgs = None
        self.entryBlock = None
        self.blocks = None
        self.procs = None # used to store information on subprocedues (from the JSR instructions)

        self.block_numberer = itertools.count(-4,-1)

    def condenseBlocks(self):
        assert not self.procs
        old = self.blocks
        # Can't do a consistency check on entry as the graph may be in an inconsistent state at this point
        # Since the purpose of this function is to prune unreachable blocks from self.blocks
        sccs = graph_util.tarjanSCC([self.entryBlock], lambda block:block.jump.getSuccessors())
        self.blocks = list(itertools.chain.from_iterable(list(map(reversed, sccs[::-1]))))

        assert set(self.blocks) <= set(old)
        if len(self.blocks) < len(old):
            kept = set(self.blocks)
            for block in self.blocks:
                for pair in block.predecessors[:]:
                    if pair[0] not in kept:
                        block.removePredPair(pair)
            return [b for b in old if b not in kept]
        return []

    def removeUnusedVariables(self):
        assert not self.procs
        roots = [x for x in self.inputArgs if x is not None]
        for block in self.blocks:
            roots += block.jump.params
            for op in block.lines:
                if op.has_side_effects:
                    roots += op.params

        reachable = graph_util.topologicalSort(roots, lambda var:(var.origin.params if var.origin else []))

        keepset = set(reachable)
        assert None not in keepset
        def filterOps(oldops):
            newops = []
            for op in oldops:
                # if any of the params is being removed due to being unreachable, we can assume the whole function can be removed
                keep = keepset.issuperset(op.params) and (op.has_side_effects or not keepset.isdisjoint(op.getOutputs()))
                if keep:
                    newops.append(op)
                    for v in op.getOutputs():
                        if v and v not in keepset:
                            op.removeOutput(v)
                else:
                    assert keepset.isdisjoint(op.getOutputs())
                    assert not op.has_side_effects
            return newops

        for block in self.blocks:
            block.phis = filterOps(block.phis)
            block.lines = filterOps(block.lines)
            block.filterVarConstraints(keepset)
        assert self._conscheck() is None

    def mergeSingleSuccessorBlocks(self):
        assert(not self.procs) # Make sure that all single jsr procs are inlined first
        assert self._conscheck() is None

        removed = set()
        for block in self.blocks:
            if block in removed:
                continue
            while isinstance(block.jump, ssa_jumps.Goto):
                jump = block.jump
                block2 = jump.getNormalSuccessors()[0]
                fromkey = block, False
                if block2.predecessors != [fromkey]:
                    break

                jump2 = block2.jump
                ucs = block.unaryConstraints
                ucs2 = block2.unaryConstraints
                replace = {phi.rval: phi.get(fromkey) for phi in block2.phis}
                for var2, var in list(replace.items()):
                    ucs[var] = constraints.join(ucs[var], ucs2.pop(var2))
                ucs.update(ucs2)

                for op in block2.lines:
                    op.replaceVars(replace)
                block.lines += block2.lines

                jump2.replaceVars(replace)
                block.jump = jump2

                # remember to update phis of blocks referring to old child!
                for successor, t in block.jump.getSuccessorPairs():
                    successor.replacePredPair((block2, t), (block, t))
                    for phi in successor.phis:
                        phi.replaceVars(replace)
                removed.add(block2)
        self.blocks = [b for b in self.blocks if b not in removed]
        assert self._conscheck() is None

    def disconnectConstantVariables(self):
        for block in self.blocks:
            for var, uc in list(block.unaryConstraints.items()):
                if var.origin is not None:
                    newval = None
                    if var.type[0] == 'int':
                        if uc.min == uc.max:
                            newval = uc.min
                    elif var.type[0] == 'obj':
                        if uc.isConstNull():
                            newval = 'null'

                    if newval is not None:
                        var.origin.removeOutput(var)
                        var.origin = None
                        var.const = newval
            block.phis = [phi for phi in block.phis if phi.rval is not None]
        assert self._conscheck() is None

    def _conscheck(self):
        '''Sanity check'''
        for block in self.blocks:
            assert block.jump is not None
            for phi in block.phis:
                assert phi.rval is None or phi.rval in block.unaryConstraints
                for k,v in list(phi.dict.items()):
                    assert v in k[0].unaryConstraints

        keys = [block.key for block in self.blocks]
        assert len(set(keys)) == len(keys)
        temp = [self.entryBlock]
        for proc in self.procs:
            temp += [proc.retblock]
            temp += proc.jsrblocks
        assert len(set(temp)) == len(temp)

    def copyPropagation(self):
        # Loop aware copy propagation
        assert not self.procs
        assert self._conscheck() is None

        # The goal is to propagate constants that would never be inferred pessimistically
        # due to the prescence of loops. Variables that aren't derived from a constant or phi
        # are treated as opaque and variables are processed by SCC in topological order.
        # For each scc, we can infer that it is the meet of all inputs that come from variables
        # in different sccs that come before it in topological order, thus ignoring variables
        # in the current scc (the loop problem).
        v2b = {}
        assigns = collections.OrderedDict()
        for block in self.blocks:
            for var in block.unaryConstraints:
                v2b[var] = block
            for phi in block.phis:
                assigns[phi.rval] = list(map(phi.get, block.predecessors))

        UCs = {}
        sccs = graph_util.tarjanSCC(assigns, lambda v:assigns.get(v, []))
        for scc in sccs:
            if all(var in assigns for var in scc):
                invars = sum(list(map(assigns.get, scc)), [])
                inputs = [UCs[invar] for invar in invars if invar in UCs]
                assert inputs
                uc = constraints.meet(*inputs)

                for var in scc:
                    old = v2b[var].unaryConstraints[var]
                    new = constraints.join(uc, old) or old # temporary hack
                    v2b[var].unaryConstraints[var] = UCs[var] = new
            else:
                # There is a root in this scc, so we can't do anything
                for var in scc:
                    UCs[var] = v2b[var].unaryConstraints[var]

        assert self._conscheck() is None

    def abstractInterpert(self):
        # Sparse conditional constant propagation and type inference
        assert not self.procs
        assert self._conscheck() is None

        visit_counts = collections.defaultdict(int)
        dirty_phis = set(itertools.chain.from_iterable(block.phis for block in self.blocks))
        while dirty_phis:
            for block in self.blocks:
                assert block in self.blocks
                UCs = block.unaryConstraints
                assert None not in list(UCs.values())
                dirty = visit_counts[block] == 0
                for phi in block.phis:
                    if phi in dirty_phis:
                        dirty_phis.remove(phi)
                        inputs = [key[0].unaryConstraints[phi.get(key)] for key in block.predecessors]
                        out = constraints.meet(*inputs)
                        old = UCs[phi.rval]
                        UCs[phi.rval] = out = constraints.join(old, out)
                        dirty = dirty or out != old
                        assert out

                if not dirty or visit_counts[block] >= 5:
                    continue
                visit_counts[block] += 1

                must_throw = False
                dirty_vars = set()
                last_line = block.lines[-1] if block.lines else None # Keep reference handy to exception phi, if any
                for i, op in enumerate(block.lines):
                    if hasattr(op, 'propagateConstraints'):
                        output_vars = op.getOutputs()
                        inputs = [UCs[var] for var in op.params]
                        assert None not in inputs
                        output_info = op.propagateConstraints(*inputs)

                        for var, out in zip(output_vars, [output_info.rval, output_info.eval]):
                            if var is None:
                                continue
                            old = UCs[var]
                            UCs[var] = out = constraints.join(old, out)
                            if out is None:
                                if var is op.outException:
                                    assert isinstance(last_line, ssa_ops.ExceptionPhi)
                                    last_line.params.remove(var)
                                op.removeOutput(var) # Note, this must be done after the op.outException check!
                                del UCs[var]
                            elif out != old:
                                dirty_vars.add(var)

                        if output_info.must_throw:
                            must_throw = True
                            # Remove all code after this in the basic block and adjust exception code
                            # at end as appropriate
                            assert isinstance(last_line, ssa_ops.ExceptionPhi)
                            assert i < len(block.lines) and op.outException
                            removed = block.lines[i+1:-1]
                            block.lines = block.lines[:i+1] + [last_line]
                            for op2 in removed:
                                if op2.outException:
                                    last_line.params.remove(op2.outException)
                                for var in op2.getOutputs():
                                    if var is not None:
                                        del UCs[var]
                            break

                # now handle end of block
                if isinstance(last_line, ssa_ops.ExceptionPhi):
                    inputs = list(map(UCs.get, last_line.params))
                    out = constraints.meet(*inputs)
                    old = UCs[last_line.outException]
                    assert out is None or not out.null
                    UCs[last_line.outException] = out = constraints.join(old, out)
                    if out is None:
                        del UCs[last_line.outException]
                        block.lines.pop()
                    elif out != old:
                        dirty_vars.add(last_line.outException)

                # prune jumps
                dobreak = False
                if hasattr(block.jump, 'constrainJumps'):
                    assert block.jump.params
                    oldEdges = block.jump.getSuccessorPairs()
                    inputs = list(map(UCs.get, block.jump.params))
                    block.jump = block.jump.constrainJumps(*inputs)
                    # No exception case ordinarily won't be pruned, so we have to handle it explicitly
                    if must_throw and isinstance(block.jump, ssa_jumps.OnException):
                        if block.jump.getNormalSuccessors(): # make sure it wasn't already pruned
                            fallthrough = block.jump.getNormalSuccessors()[0]
                            block.jump = block.jump.reduceSuccessors([(fallthrough, False)])

                    newEdges = block.jump.getSuccessorPairs()
                    if newEdges != oldEdges:
                        pruned = [x for x in oldEdges if x not in newEdges]
                        for (child,t) in pruned:
                            child.removePredPair((block,t))

                        removed_blocks = self.condenseBlocks()
                        # In case where no blocks were removed, self.blocks will possibly be in a different
                        # order than the version of self.blocks we are iterating over, but it still has the
                        # same contents, so this should be safe. If blocks were removed, we break out of the
                        # list and restart to avoid the possibility of processing an unreachable block.
                        dobreak = len(removed_blocks) > 0
                        for removed in removed_blocks:
                            for phi in removed.phis:
                                dirty_phis.discard(phi)

                # update dirty set
                for child, t in block.jump.getSuccessorPairs():
                    assert child in self.blocks
                    for phi in child.phis:
                        if phi.get((block, t)) in dirty_vars:
                            dirty_phis.add(phi)
                if dobreak:
                    break

        # Try to turn switches into if statements - note that this may
        # introduce a new variable and this modify block.unaryConstraints
        # However, it won't change the control flow graph structure
        for block in self.blocks:
            if isinstance(block.jump, ssa_jumps.Switch):
                block.jump = block.jump.simplifyToIf(block)

    def simplifyThrows(self):
        # Try to turn throws into gotos where possible. This primarily helps with certain patterns of try-with-resources
        # To do this, the exception must be known to be non null and there must be only one target that can catch it
        # As a heuristic, we also restrict it to cases where every predecessor of the target can be converted
        candidates = collections.defaultdict(list)
        for block in self.blocks:
            if not isinstance(block.jump, ssa_jumps.OnException) or len(block.jump.getSuccessorPairs()) != 1:
                continue
            if len(block.lines[-1].params) != 1 or not isinstance(block.lines[-2], ssa_ops.Throw):
                continue
            if block.unaryConstraints[block.lines[-2].params[0]].null:
                continue

            candidates[block.jump.getExceptSuccessors()[0]].append(block)

        for child in self.blocks:
            if not candidates[child] or len(candidates[child]) < len(child.predecessors):
                continue

            for parent in candidates[child]:
                ephi = parent.lines.pop()
                throw_op = parent.lines.pop()

                var1 = throw_op.params[0]
                var2 = throw_op.outException
                assert ephi.params == [var2]
                var3 = ephi.outException
                assert parent.jump.params[0] == var3

                for phi in child.phis:
                    phi.replaceVars({var3: var1})
                child.replacePredPair((parent, True), (parent, False))

                del parent.unaryConstraints[var2]
                del parent.unaryConstraints[var3]
                parent.jump = ssa_jumps.Goto(self, child)

    def simplifyCatchIgnored(self):
        # When there is a single throwing instruction, which is garuenteed to throw, has a single handler, and
        # the caught exception is unused, turn it into a goto. This simplifies a pattern used by some obfuscators
        # that do stuff like try{new int[-1];} catch(Exception e) {...}
        candidates = collections.defaultdict(list)
        for block in self.blocks:
            if not isinstance(block.jump, ssa_jumps.OnException) or len(block.jump.getSuccessorPairs()) != 1:
                continue
            if len(block.lines[-1].params) != 1:
                continue
            candidates[block.jump.getExceptSuccessors()[0]].append(block)

        for child in self.blocks:
            if not candidates[child] or len(candidates[child]) < len(child.predecessors):
                continue

            # Make sure caught exception is unused
            temp = candidates[child][0].lines[-1].outException
            if any(temp in phi.params for phi in child.phis):
                continue

            for parent in candidates[child]:
                ephi = parent.lines.pop()
                throw_op = parent.lines.pop()
                del parent.unaryConstraints[throw_op.outException]
                del parent.unaryConstraints[ephi.outException]
                child.replacePredPair((parent, True), (parent, False))
                parent.jump = ssa_jumps.Goto(self, child)

    # Subprocedure stuff #####################################################
    def _newBlockFrom(self, block):
        b = BasicBlock(next(self.block_numberer))
        self.blocks.append(b)
        return b

    def _copyVar(self, var, vard=None):
        v = copy.copy(var)
        v.name = v.origin = None # TODO - generate new names?
        if vard is not None:
            vard[var] = v
        return v

    def _region(self, proc):
        # Find the set of blocks 'in' a subprocedure, i.e. those reachable from the target that can reach the ret block
        region = graph_util.topologicalSort([proc.retblock], lambda block:[] if block == proc.target else [b for b,t in block.predecessors])
        temp = set(region)
        assert self.entryBlock not in temp and proc.target in temp and temp.isdisjoint(proc.jsrblocks)
        return region

    def _duplicateBlocks(self, region, excludedPreds):
        # Duplicate a region of blocks. All inedges will be redirected to the new blocks
        # except for those from excludedPreds
        excludedPreds = excludedPreds | set(region)
        outsideBlocks = [b for b in self.blocks if b not in excludedPreds]

        blockd, vard = {}, {}
        for oldb in region:
            block = blockd[oldb] = self._newBlockFrom(oldb)
            block.unaryConstraints = {self._copyVar(k, vard):v for k, v in list(oldb.unaryConstraints.items())}
            block.phis = [ssa_ops.Phi(block, vard[oldphi.rval]) for oldphi in oldb.phis]

            for op in oldb.lines:
                new = copy.copy(op)
                new.replaceVars(vard)
                new.replaceOutVars(vard)
                assert new.getOutputs().count(None) == op.getOutputs().count(None)
                for outv in new.getOutputs():
                    if outv is not None:
                        assert outv.origin is None
                        outv.origin = new
                block.lines.append(new)

            assert set(vard).issuperset(oldb.jump.params)
            block.jump = oldb.jump.clone()
            block.jump.replaceVars(vard)

            # Fix up blocks outside the region that jump into the region.
            for key in oldb.predecessors[:]:
                pred = key[0]
                if pred not in excludedPreds:
                    for phi1, phi2 in zip(oldb.phis, block.phis):
                        phi2.add(key, phi1.get(key))
                        del phi1.dict[key]
                    oldb.predecessors.remove(key)
                    block.predecessors.append(key)

        # fix up jump targets of newly created blocks
        for oldb, block in list(blockd.items()):
            block.jump.replaceBlocks(blockd)
            for suc, t in block.jump.getSuccessorPairs():
                suc.predecessors.append((block, t))

        # update the jump targets of predecessor blocks
        for block in outsideBlocks:
            block.jump.replaceBlocks(blockd)

        for old, new in list(vard.items()):
            assert type(old.origin) == type(new.origin)

        # Fill in phi args in successors of new blocks
        for oldb, block in list(blockd.items()):
            for oldc, t in oldb.jump.getSuccessorPairs():
                child = blockd.get(oldc, oldc)
                assert len(child.phis) == len(oldc.phis)
                for phi1, phi2 in zip(oldc.phis, child.phis):
                    phi2.add((block, t), vard[phi1.get((oldb, t))])

        assert self._conscheck() is None
        return blockd

    def _splitSubProc(self, proc):
        # Splits a proc into two, with one callsite using the new proc instead
        # this involves duplicating the body of the procedure
        # the new proc is appended to the list of procs so it can work properly
        # with the stack processing in inlineSubprocs
        assert len(proc.jsrblocks) > 1
        target, retblock = proc.target, proc.retblock
        region = self._region(proc)

        split_jsrs = [proc.jsrblocks.pop()]
        blockd = self._duplicateBlocks(region, set(proc.jsrblocks))

        newproc = subproc.ProcInfo(blockd[proc.retblock], blockd[proc.target])
        newproc.jsrblocks = split_jsrs
        # Sanity check
        for temp in self.procs + [newproc]:
            for jsr in temp.jsrblocks:
                assert jsr.jump.target == temp.target
        return newproc

    def _inlineSubProc(self, proc):
        # Inline a proc with single callsite inplace
        assert len(proc.jsrblocks) == 1
        target, retblock = proc.target, proc.retblock
        region = self._region(proc)

        jsrblock = proc.jsrblocks[0]
        jsrop = jsrblock.jump
        ftblock = jsrop.fallthrough

        # first we find any vars that bypass the proc since we have to pass them through the new blocks
        skipvars = [phi.get((jsrblock, False)) for phi in ftblock.phis]
        skipvars = [var for var in skipvars if var.origin is not jsrop]

        svarcopy = {(var, block):self._copyVar(var) for var, block in itertools.product(skipvars, region)}
        for var, block in itertools.product(skipvars, region):
            # Create a new phi for the passed through var for this block
            rval = svarcopy[var, block]
            phi = ssa_ops.Phi(block, rval)
            block.phis.append(phi)
            block.unaryConstraints[rval] = jsrblock.unaryConstraints[var]

            for key in block.predecessors:
                if key == (jsrblock, False):
                    phi.add(key, var)
                else:
                    phi.add(key, svarcopy[var, key[0]])

        outreplace = {jv:rv for jv, rv in zip(jsrblock.jump.output.stack, retblock.jump.input.stack) if jv is not None}
        outreplace.update({jv:retblock.jump.input.locals[i] for i, jv in list(jsrblock.jump.output.locals.items()) if jv is not None})
        for var in outreplace: # don't need jsrop's out vars anymore
            del jsrblock.unaryConstraints[var]

        for var in skipvars:
            outreplace[var] = svarcopy[var, retblock]
        jsrblock.jump = ssa_jumps.Goto(self, target)
        retblock.jump = ssa_jumps.Goto(self, ftblock)

        ftblock.replacePredPair((jsrblock, False), (retblock, False))
        for phi in ftblock.phis:
            phi.replaceVars(outreplace)

    def inlineSubprocs(self):
        assert self._conscheck() is None
        assert self.procs

        # establish DAG of subproc callstacks if we're doing nontrivial inlining, since we can only inline leaf procs
        regions = {proc:frozenset(self._region(proc)) for proc in self.procs}
        parents = {proc:[] for proc in self.procs}
        for x,y in itertools.product(self.procs, repeat=2):
            if not regions[y].isdisjoint(x.jsrblocks):
                parents[x].append(y)

        self.procs = graph_util.topologicalSort(self.procs, parents.get)
        if any(parents.values()):
            print('Warning, nesting subprocedures detected! This method may take a long time to decompile.')
        print('Subprocedures for', self.code.method.name + ':', self.procs)

        # now inline the procs
        while self.procs:
            proc = self.procs.pop()
            while len(proc.jsrblocks) > 1:
                print('splitting', proc)
                # push new subproc onto stack
                self.procs.append(self._splitSubProc(proc))
                assert self._conscheck() is None
            # When a subprocedure has only one call point, it can just be inlined instead of splitted
            print('inlining', proc)
            self._inlineSubProc(proc)
            assert self._conscheck() is None
    ##########################################################################
    def splitDualInedges(self):
        # Split any blocks that have both normal and exceptional in edges
        assert not self.procs
        for block in self.blocks[:]:
            if block is self.entryBlock:
                continue
            types = set(zip(*block.predecessors)[1])
            if len(types) <= 1:
                continue
            assert not isinstance(block.jump, (ssa_jumps.Return, ssa_jumps.Rethrow))

            new = self._newBlockFrom(block)
            print('Splitting', block, '->', new)
            # first fix up CFG edges
            badpreds = [t for t in block.predecessors if t[1]]
            new.predecessors = badpreds
            for t in badpreds:
                block.predecessors.remove(t)

            for pred, _ in badpreds:
                assert isinstance(pred.jump, ssa_jumps.OnException)
                pred.jump.replaceExceptTarget(block, new)

            new.jump = ssa_jumps.Goto(self, block)
            block.predecessors.append((new, False))

            # fix up variables
            new.phis = []
            new.unaryConstraints = {}
            for phi in block.phis:
                newrval = self._copyVar(phi.rval)
                new.unaryConstraints[newrval] = block.unaryConstraints[phi.rval]
                newphi = ssa_ops.Phi(new, newrval)
                new.phis.append(newphi)

                for t in badpreds:
                    arg = phi.get(t)
                    phi.delete(t)
                    newphi.add(t, arg)
                phi.add((new, False), newrval)
        assert self._conscheck() is None

    def fixLoops(self):
        assert not self.procs
        todo = self.blocks[:]
        while todo:
            newtodo = []
            temp = set(todo)
            sccs = graph_util.tarjanSCC(todo, lambda block:[x for x,t in block.predecessors if x in temp])

            for scc in sccs:
                if len(scc) <= 1:
                    continue

                scc_pair_set = {(x, False) for x in scc} | {(x, True) for x in scc}
                entries = [n for n in scc if not scc_pair_set.issuperset(n.predecessors)]

                if len(entries) <= 1:
                    head = entries[0]
                else:
                    # if more than one entry point into the loop, we have to choose one as the head and duplicate the rest
                    print('Warning, multiple entry point loop detected. Generated code may be extremely large', end=' ')
                    print('({} entry points, {} blocks)'.format(len(entries), len(scc)))
                    def loopSuccessors(head, block):
                        if block == head:
                            return []
                        return [x for x in block.jump.getSuccessors() if (x, False) in scc_pair_set]

                    reaches = [(n, graph_util.topologicalSort(entries, functools.partial(loopSuccessors, n))) for n in scc]
                    for head, reachable in reaches:
                        reachable.remove(head)

                    head, reachable = min(reaches, key=lambda t:(len(t[1]), -len(t[0].predecessors)))
                    assert head not in reachable
                    print('Duplicating {} nodes'.format(len(reachable)))
                    blockd = self._duplicateBlocks(reachable, set(scc) - set(reachable))
                    newtodo += list(map(blockd.get, reachable))
                newtodo.extend(scc)
                newtodo.remove(head)
            todo = newtodo
        assert self._conscheck() is None

    # Functions called by children ###########################################
    # assign variable names for debugging
    varnum = collections.defaultdict(itertools.count)
    def makeVariable(self, *args, **kwargs):
        # Note: Make sure this doesn't hold on to created variables in any way,
        # since this func may be called for temporary results that are discarded
        var = SSA_Variable(*args, **kwargs)
        # pref = args[0][0][0].replace('o','a')
        # var.name = pref + str(next(self.varnum[pref]))
        return var

    def setObjVarData(self, var, vtype, initMap):
        vtype2 = initMap.get(vtype, vtype)
        tt = objtypes.verifierToSynthetic(vtype2)
        assert var.decltype is None or var.decltype == tt
        var.decltype = tt
        # if uninitialized, record the offset of originating new instruction for later
        if vtype.tag == '.new':
            assert var.uninit_orig_num is None or var.uninit_orig_num == vtype.extra
            var.uninit_orig_num = vtype.extra

    def makeVarFromVtype(self, vtype, initMap):
        vtype2 = initMap.get(vtype, vtype)
        type_ = verifierToSSAType(vtype2)
        if type_ is not None:
            var = self.makeVariable(type_)
            if type_ == SSA_OBJECT:
                self.setObjVarData(var, vtype, initMap)
            return var
        return None

    def getConstPoolArgs(self, index):
        return self.class_.cpool.getArgs(index)

    def getConstPoolType(self, index):
        return self.class_.cpool.getType(index)

def ssaFromVerified(code, iNodes, opts):
    method = code.method
    inputTypes, returnTypes = parseUnboundMethodDescriptor(method.descriptor, method.class_.name, method.static)

    parent = SSA_Graph(code)
    data = blockmaker.BlockMaker(parent, iNodes, inputTypes, returnTypes, code.except_raw, opts=opts)

    parent.blocks = blocks = data.blocks
    parent.entryBlock = data.entryBlock
    parent.inputArgs = data.inputArgs
    assert parent.entryBlock in blocks

    # create subproc info
    procd = {block.jump.target: subproc.ProcInfo(block, block.jump.target) for block in blocks if isinstance(block.jump, subproc.DummyRet)}
    for block in blocks:
        if isinstance(block.jump, subproc.ProcCallOp):
            procd[block.jump.target].jsrblocks.append(block)
    parent.procs = sorted(list(procd.values()), key=lambda p:p.target.key)

    # Intern constraints to save a bit of memory for long methods
    def makeConstraint(var, _cache={}):
        key = var.type, var.const, var.decltype, var.uninit_orig_num is None
        try:
            return _cache[key]
        except KeyError:
            _cache[key] = temp = constraints.fromVariable(parent.env, var)
            return temp

    # create unary constraints for each variable
    for block in blocks:
        bvars = []
        if isinstance(block.jump, subproc.ProcCallOp):
            bvars += block.jump.flatOutput()
        # entry block has no phis
        if block is parent.entryBlock:
            bvars += parent.inputArgs

        bvars = [v for v in bvars if v is not None]
        bvars += [phi.rval for phi in block.phis]
        for op in block.lines:
            bvars += op.params
            bvars += [x for x in op.getOutputs() if x is not None]
        bvars += block.jump.params

        for suc, t in block.jump.getSuccessorPairs():
            for phi in suc.phis:
                bvars.append(phi.get((block, t)))
        assert None not in bvars
        # Note that makeConstraint can indirectly cause class loading
        block.unaryConstraints = {var:makeConstraint(var) for var in bvars}

    parent._conscheck()
    return parent



_pairs = [
    ('public', 0x0001),
    ('private', 0x0002),
    ('protected', 0x0004),
    ('static', 0x0008),
    ('final', 0x0010),
    ('super', 0x0020),
    # todo - order attributes properly by context
    ('transitive', 0x0020),
    ('open', 0x0020),
    ('synchronized', 0x0020),
    ('volatile', 0x0040),
    ('static_phase', 0x0040),
    ('bridge', 0x0040),
    ('transient', 0x0080),
    ('varargs', 0x0080),
    ('native', 0x0100),
    ('interface', 0x0200),
    ('abstract', 0x0400),
    ('strict', 0x0800),
    ('synthetic', 0x1000),
    ('annotation', 0x2000),
    ('enum', 0x4000),
    ('module', 0x8000),
    ('mandated', 0x8000),
]

FLAGS = dict(_pairs)
# Treat strictfp as flag too to reduce confusion
FLAGS['strictfp'] = FLAGS['strict']

def _make_dict(priority):
    d = {v:k for k,v in reversed(_pairs)}
    # ensure that the specified flags have priority
    for flag in priority.split():
        d[FLAGS[flag]] = flag
    return d

RFLAGS_CLASS = _make_dict('super module')
RFLAGS_FIELD = _make_dict('volatile transient')
RFLAGS_METHOD = _make_dict('synchronized bridge varargs')
RFLAGS_MOD_REQUIRES = _make_dict('transitive static_phase mandated')
RFLAGS_MOD_OTHER = _make_dict('open mandated')



from .. import opnames
from ..namegen import LabelGen
from ..ssa import objtypes, ssa_jumps, ssa_ops, ssa_types
from ..verifier.descriptors import parseFieldDescriptor, parseMethodDescriptor

from . import ast
from .setree import SEBlockItem, SEIf, SEScope, SESwitch, SETry, SEWhile

# prefixes for name generation
_prefix_map = {objtypes.IntTT:'i', objtypes.LongTT:'j',
            objtypes.FloatTT:'f', objtypes.DoubleTT:'d',
            objtypes.BoolTT:'b', objtypes.StringTT:'s'}

_ssaToTT = {ssa_types.SSA_INT:objtypes.IntTT, ssa_types.SSA_LONG:objtypes.LongTT,
            ssa_types.SSA_FLOAT:objtypes.FloatTT, ssa_types.SSA_DOUBLE:objtypes.DoubleTT}
class VarInfo(object):
    def __init__(self, method, blocks, namegen):
        self.env = method.class_.env
        self.labelgen = LabelGen().next

        returnTypes = parseMethodDescriptor(method.descriptor, unsynthesize=False)[-1]
        self.return_tt = objtypes.verifierToSynthetic(returnTypes[0]) if returnTypes else None
        self.clsname = method.class_.name
        self._namegen = namegen

        self._uninit_vars = {}
        self._vars = {}
        self._tts = {}
        for block in blocks:
            for var, uc in block.unaryConstraints.items():
                if var.type == ssa_types.SSA_OBJECT:
                    tt = uc.getSingleTType() # temp hack
                    if uc.types.isBoolOrByteArray():
                        tt = objtypes.TypeTT(objtypes.BExpr, objtypes.dim(tt)+1)
                        # assert (objtypes.BoolTT[0], tt[1]) in uc.types.exact
                else:
                    tt = _ssaToTT[var.type]
                self._tts[var] = tt

    def _nameCallback(self, expr):
        prefix = _prefix_map.get(expr.dtype, 'a')
        return self._namegen.getPrefix(prefix)

    def _newVar(self, var, num, isCast):
        tt = self._tts[var]
        if var.const is not None and not isCast:
            return ast.Literal(tt, var.const)

        if var.name:
            # important to not add num when it is 0, since we currently
            # use var names to force 'this'
            temp = '{}_{}'.format(var.name, num) if num else var.name
            if isCast:
                temp += 'c'
            namefunc = lambda expr:temp
        else:
            namefunc = self._nameCallback

        result = ast.Local(tt, namefunc)
        # merge all variables of uninitialized type to simplify fixObjectCreations in javamethod.py
        if var.uninit_orig_num is not None and not isCast:
            result = self._uninit_vars.setdefault(var.uninit_orig_num, result)
        return result

    def var(self, node, var, isCast=False):
        key = node, var, isCast
        try:
            return self._vars[key]
        except KeyError:
            new = self._newVar(key[1], key[0].num, key[2])
            self._vars[key] = new
            return new

    def customVar(self, tt, prefix): # for use with ignored exceptions
        namefunc = lambda expr: self._namegen.getPrefix(prefix)
        return ast.Local(tt, namefunc)

#########################################################################################
_math_types = (ssa_ops.IAdd, ssa_ops.IDiv, ssa_ops.IMul, ssa_ops.IRem, ssa_ops.ISub)
_math_types += (ssa_ops.IAnd, ssa_ops.IOr, ssa_ops.IShl, ssa_ops.IShr, ssa_ops.IUshr, ssa_ops.IXor)
_math_types += (ssa_ops.FAdd, ssa_ops.FDiv, ssa_ops.FMul, ssa_ops.FRem, ssa_ops.FSub)
_math_symbols = dict(zip(_math_types, '+ / * % - & | << >> >>> ^ + / * % -'.split()))
def _convertJExpr(op, getExpr, clsname):
    params = [getExpr(var) for var in op.params]
    assert None not in params
    expr = None

    # Have to do this one seperately since it isn't an expression statement
    if isinstance(op, ssa_ops.Throw):
        return ast.ThrowStatement(params[0])

    if isinstance(op, _math_types):
        opdict = _math_symbols
        expr = ast.BinaryInfix(opdict[type(op)], params)
    elif isinstance(op, ssa_ops.ArrLength):
        expr = ast.FieldAccess(params[0], 'length', objtypes.IntTT)
    elif isinstance(op, ssa_ops.ArrLoad):
        expr = ast.ArrayAccess(*params)
    elif isinstance(op, ssa_ops.ArrStore):
        expr = ast.ArrayAccess(params[0], params[1])
        expr = ast.Assignment(expr, params[2])
    elif isinstance(op, ssa_ops.CheckCast):
        expr = ast.Cast(ast.TypeName(op.target_tt), params[0])
    elif isinstance(op, ssa_ops.Convert):
        expr = ast.makeCastExpr(_ssaToTT[op.target], params[0])
    elif isinstance(op, (ssa_ops.FCmp, ssa_ops.ICmp)):
        boolt = objtypes.BoolTT
        cn1, c0, c1 = ast.Literal.N_ONE, ast.Literal.ZERO, ast.Literal.ONE

        ascend = isinstance(op, ssa_ops.ICmp) or op.NaN_val == 1
        if ascend:
            expr = ast.Ternary(ast.BinaryInfix('<',params,boolt), cn1, ast.Ternary(ast.BinaryInfix('==',params,boolt), c0, c1))
        else:
            assert op.NaN_val == -1
            expr = ast.Ternary(ast.BinaryInfix('>',params,boolt), c1, ast.Ternary(ast.BinaryInfix('==',params,boolt), c0, cn1))
    elif isinstance(op, ssa_ops.FieldAccess):
        dtype = objtypes.verifierToSynthetic(parseFieldDescriptor(op.desc, unsynthesize=False)[0])

        if op.instruction[0] in (opnames.GETSTATIC, opnames.PUTSTATIC):
            printLeft = (op.target != clsname) # Don't print classname if it is a static field in current class
            tt = objtypes.TypeTT(op.target, 0) # Doesn't handle arrays, but they don't have any fields anyway
            expr = ast.FieldAccess(ast.TypeName(tt), op.name, dtype, op, printLeft=printLeft)
        else:
            expr = ast.FieldAccess(params[0], op.name, dtype, op)

        if op.instruction[0] in (opnames.PUTFIELD, opnames.PUTSTATIC):
            expr = ast.Assignment(expr, params[-1])

    elif isinstance(op, ssa_ops.FNeg):
        expr = ast.UnaryPrefix('-', params[0])
    elif isinstance(op, ssa_ops.InstanceOf):
        args = [params[0], ast.TypeName(op.target_tt)]
        expr = ast.BinaryInfix('instanceof', args, dtype=objtypes.BoolTT)
    elif isinstance(op, ssa_ops.Invoke):
        vtypes, rettypes = parseMethodDescriptor(op.desc, unsynthesize=False)
        tt_types = objtypes.verifierToSynthetic_seq(vtypes)
        ret_type = objtypes.verifierToSynthetic(rettypes[0]) if rettypes else None
        target_tt = op.target_tt

        if objtypes.dim(target_tt) and op.name == "clone": # In Java, T[].clone returns T[] rather than Object
            ret_type = target_tt

        if op.instruction[0] == opnames.INVOKEINIT and op.isThisCtor:
            name = 'this' if (op.target == clsname) else 'super'
            expr = ast.MethodInvocation(None, name, tt_types, params[1:], op, ret_type)
        elif op.instruction[0] == opnames.INVOKESTATIC: # TODO - fix this for special super calls
            expr = ast.MethodInvocation(ast.TypeName(target_tt), op.name, [None]+tt_types, params, op, ret_type)
        else:
            expr = ast.MethodInvocation(params[0], op.name, [target_tt]+tt_types, params[1:], op, ret_type)
    elif isinstance(op, ssa_ops.InvokeDynamic):
        vtypes, rettypes = parseMethodDescriptor(op.desc, unsynthesize=False)
        ret_type = objtypes.verifierToSynthetic(rettypes[0]) if rettypes else None
        fmt = '/*invokedynamic*/'
        if ret_type is not None:
            fmt += '{{{}}}'.format(len(params))
            params.append(ast.dummyLiteral(ret_type))
        expr = ast.Dummy(fmt, params, dtype=ret_type)
    elif isinstance(op, ssa_ops.Monitor):
        fmt = '/*monexit({})*/' if op.exit else '/*monenter({})*/'
        expr = ast.Dummy(fmt, params)
    elif isinstance(op, ssa_ops.MultiNewArray):
        expr = ast.ArrayCreation(op.tt, *params)
    elif isinstance(op, ssa_ops.New):
        expr = ast.Dummy('/*<unmerged new> {}*/', [ast.TypeName(op.tt)], isNew=True)
    elif isinstance(op, ssa_ops.NewArray):
        expr = ast.ArrayCreation(op.tt, params[0])
    elif isinstance(op, ssa_ops.Truncate):
        tt = {(True,16): objtypes.ShortTT, (False,16): objtypes.CharTT, (True,8): objtypes.ByteTT}[op.signed, op.width]
        expr = ast.Cast(ast.TypeName(tt), params[0])
    if op.rval is not None and expr:
        expr = ast.Assignment(getExpr(op.rval), expr)

    if expr is None: # Temporary hack
        if isinstance(op, (ssa_ops.TryReturn, ssa_ops.ExceptionPhi, ssa_ops.MagicThrow)):
            return None # Don't print out anything
    return ast.ExpressionStatement(expr)

#########################################################################################
def _createASTBlock(info, endk, node):
    getExpr = lambda var: info.var(node, var)
    op2expr = lambda op: _convertJExpr(op, getExpr, info.clsname)

    block = node.block
    if block is not None:
        split_ind = 0
        if isinstance(block.jump, ssa_jumps.OnException):
            # find index of first throwing instruction, so we can insert eassigns before it later
            assert isinstance(block.lines[-1], ssa_ops.ExceptionPhi)
            split_ind = block.lines.index(block.lines[-1].params[0].origin)

        lines_before = filter(None, map(op2expr, block.lines[:split_ind]))
        lines_after = filter(None, map(op2expr, block.lines[split_ind:]))
    else:
        lines_before, lines_after = [], []

    # Kind of hackish: If the block ends in a cast and hence it is not known to always
    # succeed, assign the results of the cast rather than passing through the variable
    # unchanged. The cast will actually be second to last in block.lines due to the ephi
    outreplace = {}
    if block and len(block.lines) >= 2:
        temp_op = block.lines[-2]
        if lines_after and isinstance(temp_op, ssa_ops.CheckCast):
            assert isinstance(lines_after[-1].expr, ast.Cast)
            var = temp_op.params[0]
            cexpr = lines_after[-1].expr
            lhs = info.var(node, var, True)
            assert lhs != cexpr.params[1]
            lines_after[-1].expr = ast.Assignment(lhs, cexpr)
            nvar = outreplace[var] = lines_after[-1].expr.params[0]
            nvar.dtype = cexpr.dtype

    eassigns = []
    nassigns = []
    for n2 in node.successors:
        assert (n2 in node.outvars) != (n2 in node.eassigns)
        if n2 in node.eassigns:
            for outv, inv in zip(node.eassigns[n2], n2.invars):
                if outv is None: # this is how we mark the thrown exception, which
                    # obviously doesn't get an explicit assignment statement
                    continue
                expr = ast.Assignment(info.var(n2, inv), info.var(node, outv))
                if expr.params[0] != expr.params[1]:
                    eassigns.append(ast.ExpressionStatement(expr))
        else:
            for outv, inv in zip(node.outvars[n2], n2.invars):
                right = outreplace.get(outv, info.var(node, outv))
                expr = ast.Assignment(info.var(n2, inv), right)
                if expr.params[0] != expr.params[1]:
                    nassigns.append(ast.ExpressionStatement(expr))

    # Need to put exception assignments before first throwing statement
    # While normal assignments must come last as they may depend on it
    statements = lines_before + eassigns + lines_after + nassigns

    norm_successors = node.normalSuccessors()
    jump = None if block is None else block.jump
    if isinstance(jump, (ssa_jumps.Rethrow, ssa_jumps.Return)):
        assert not norm_successors
        assert not node.eassigns and not node.outvars
        if isinstance(jump, ssa_jumps.Rethrow):
            param = info.var(node, jump.params[-1])
            statements.append(ast.ThrowStatement(param))
        else:
            if len(jump.params) > 0:
                param = info.var(node, jump.params[0])
                statements.append(ast.ReturnStatement(param, info.return_tt))
            else:
                statements.append(ast.ReturnStatement())
        breakKey, jumpKey = endk, None
    elif len(norm_successors) == 0:
        assert isinstance(jump, ssa_jumps.OnException)
        breakKey, jumpKey = endk, None
    elif len(norm_successors) == 1: # normal successors
        breakKey, jumpKey = endk, norm_successors[0]._key
    else: # case of if and switch jumps handled in parent scope
        assert len(norm_successors) > 1
        breakKey, jumpKey = endk, endk

    new = ast.StatementBlock(info.labelgen, node._key, breakKey, statements, jumpKey)
    assert None not in statements
    return new

_cmp_strs = dict(zip(('eq','ne','lt','ge','gt','le'), "== != < >= > <=".split()))
def _createASTSub(info, current, ftitem, forceUnlabled=False):
    begink = current.entryBlock._key
    endk = ftitem.entryBlock._key if ftitem is not None else None

    if isinstance(current, SEBlockItem):
        return _createASTBlock(info, endk, current.node)
    elif isinstance(current, SEScope):
        ftitems = current.items[1:] + [ftitem]
        parts = [_createASTSub(info, item, newft) for item, newft in zip(current.items, ftitems)]
        return ast.StatementBlock(info.labelgen, begink, endk, parts, endk, labelable=(not forceUnlabled))
    elif isinstance(current, SEWhile):
        parts = [_createASTSub(info, scope, current, True) for scope in current.getScopes()]
        return ast.WhileStatement(info.labelgen, begink, endk, tuple(parts))
    elif isinstance(current, SETry):
        assert len(current.getScopes()) == 2
        parts = [_createASTSub(info, scope, ftitem, True) for scope in current.getScopes()]
        catchnode = current.getScopes()[-1].entryBlock
        declt = ast.CatchTypeNames(info.env, current.toptts)

        if current.catchvar is None: # exception is ignored and hence not referred to by the graph, so we need to make our own
            catchvar = info.customVar(declt.dtype, 'ignoredException')
        else:
            catchvar = info.var(catchnode, current.catchvar)
        decl = ast.VariableDeclarator(declt, catchvar)
        pairs = [(decl, parts[1])]
        return ast.TryStatement(info.labelgen, begink, endk, parts[0], pairs)

    # Create a fake key to represent the beginning of the conditional statement itself
    # doesn't matter what it is as long as it's unique
    midk = begink + (-1,)
    node = current.head.node
    jump = node.block.jump

    if isinstance(current, SEIf):
        parts = [_createASTSub(info, scope, ftitem, True) for scope in current.getScopes()]
        cmp_str = _cmp_strs[jump.cmp]
        exprs = [info.var(node, var) for var in jump.params]
        ifexpr = ast.BinaryInfix(cmp_str, exprs, objtypes.BoolTT)
        new = ast.IfStatement(info.labelgen, midk, endk, ifexpr, tuple(parts))

    elif isinstance(current, SESwitch):
        ftitems = current.ordered[1:] + [ftitem]
        parts = [_createASTSub(info, item, newft, True) for item, newft in zip(current.ordered, ftitems)]
        for part in parts:
            part.breakKey = endk # createSub will assume break should be ft, which isn't the case with switch statements

        expr = info.var(node, jump.params[0])
        pairs = zip(current.ordered_keysets, parts)
        new = ast.SwitchStatement(info.labelgen, midk, endk, expr, pairs)

    # bundle head and if together so we can return as single statement
    headscope = _createASTBlock(info, midk, node)
    assert headscope.jumpKey is midk
    return ast.StatementBlock(info.labelgen, begink, endk, [headscope, new], endk)

def createAST(method, ssagraph, seroot, namegen):
    info = VarInfo(method, ssagraph.blocks, namegen)
    astroot = _createASTSub(info, seroot, None)
    return astroot, info



class ValueType(object):
    '''Define _key() and inherit from this class to implement comparison and hashing'''
    # def __init__(self, *args, **kwargs): super(ValueType, self).__init__(*args, **kwargs)
    def __eq__(self, other): return type(self) == type(other) and self._key() == other._key()
    def __ne__(self, other): return type(self) != type(other) or self._key() != other._key()
    def __hash__(self): return hash(self._key())   



from ..mixin import ValueType

class IntConstraint(ValueType):
    __slots__ = "width min max".split()

    def __init__(self, width, min_, max_):
        self.width = width
        self.min = min_
        self.max = max_
        # self.isBot = (-min_ == max_+1 == (1<<width)//2)

    @staticmethod
    def range(width, min_, max_):
        if min_ > max_:
            return None
        return IntConstraint(width, min_, max_)

    @staticmethod
    def const(width, val):
        return IntConstraint(width, val, val)

    @staticmethod
    def bot(width):
        return IntConstraint(width, -1<<(width-1), (1<<(width-1))-1)

    def _key(self): return self.min, self.max

    def join(*cons):
        xmin = max(c.min for c in cons)
        xmax = min(c.max for c in cons)
        if xmin > xmax:
            return None
        res = IntConstraint(cons[0].width, xmin, xmax)
        return cons[0] if cons[0] == res else res

    def meet(*cons):
        xmin = min(c.min for c in cons)
        xmax = max(c.max for c in cons)
        return IntConstraint(cons[0].width, xmin, xmax)

    def __str__(self):   # pragma: no cover
        t = 'Int' if self.width == 32 else 'Long'
        if self.min == self.max:
            return '{}({})'.format(t, self.min)
        elif self == self.bot(self.width):
            return t
        return '{}({}, {})'.format(t, self.min, self.max)
    __repr__ = __str__



import itertools
import math

from ..ssa import objtypes

from . import visitor
from .stringescape import escapeString

# Explicitly cast parameters to the desired type in order to avoid potential issues with overloaded methods
ALWAYS_CAST_PARAMS = 1

class VariableDeclarator(object):
    def __init__(self, typename, identifier): self.typename = typename; self.local = identifier

    def print_(self, printer, print_):
        return '{} {}'.format(print_(self.typename), print_(self.local))

    def tree(self, printer, tree): return [tree(self.typename), tree(self.local)]

#############################################################################################################################################

class JavaStatement(object):
    expr = None # provide default for subclasses that don't have an expression
    def getScopes(self): return ()

    def fixLiterals(self):
        if self.expr is not None:
            self.expr = self.expr.fixLiterals()

    def addCastsAndParens(self, env):
        if self.expr is not None:
            self.expr.addCasts(env)
            self.expr.addParens()

class ExpressionStatement(JavaStatement):
    def __init__(self, expr):
        self.expr = expr
        assert expr is not None

    def print_(self, printer, print_): return print_(self.expr) + ';'
    def tree(self, printer, tree): return [self.__class__.__name__, tree(self.expr)]

class LocalDeclarationStatement(JavaStatement):
    def __init__(self, decl, expr=None):
        self.decl = decl
        self.expr = expr

    def print_(self, printer, print_):
        if self.expr is not None:
            return '{} = {};'.format(print_(self.decl), print_(self.expr))
        return print_(self.decl) + ';'

    def tree(self, printer, tree): return [self.__class__.__name__, tree(self.expr), tree(self.decl)]

    def addCastsAndParens(self, env):
        if self.expr is not None:
            self.expr.addCasts(env)

            if not isJavaAssignable(env, self.expr.dtype, self.decl.typename.tt):
                self.expr = makeCastExpr(self.decl.typename.tt, self.expr, fixEnv=env)
            self.expr.addParens()

class ReturnStatement(JavaStatement):
    def __init__(self, expr=None, tt=None):
        self.expr = expr
        self.tt = tt

    def print_(self, printer, print_): return 'return {};'.format(print_(self.expr)) if self.expr is not None else 'return;'
    def tree(self, printer, tree): return [self.__class__.__name__, tree(self.expr)]

    def addCastsAndParens(self, env):
        if self.expr is not None:
            self.expr.addCasts(env)
            if not isJavaAssignable(env, self.expr.dtype, self.tt):
                self.expr = makeCastExpr(self.tt, self.expr, fixEnv=env)
            self.expr.addParens()

class ThrowStatement(JavaStatement):
    def __init__(self, expr):
        self.expr = expr
    def print_(self, printer, print_): return 'throw {};'.format(print_(self.expr))
    def tree(self, printer, tree): return [self.__class__.__name__, tree(self.expr)]

class JumpStatement(JavaStatement):
    def __init__(self, target, isFront):
        self.label = target.getLabel() if target is not None else None
        self.keyword = 'continue' if isFront else 'break'

    def print_(self, printer, print_):
        label = ' ' + self.label if self.label is not None else ''
        return self.keyword + label + ';'

    def tree(self, printer, tree): return [self.__class__.__name__, self.keyword, self.label]

# Compound Statements
sbcount = itertools.count()
class LazyLabelBase(JavaStatement):
    # Jumps are represented by arbitrary 'keys', currently just the key of the
    # original proxy node. Each item has a continueKey and a breakKey representing
    # the beginning and the point just past the end respectively. breakKey may be
    # None if this item appears at the end of the function and there is nothing after it.
    # Statement blocks have a jump key representing where it jumps to if any. This
    # may be None if the jump is unreachable (such as if there is a throw or return)
    def __init__(self, labelfunc, begink, endk):
        self.label, self.func = None, labelfunc
        self.continueKey = begink
        self.breakKey = endk
        self.id = next(sbcount) # For debugging purposes

    def getLabel(self):
        if self.label is None:
            self.label = self.func() # Not a bound function!
        return self.label

    def getLabelPrefix(self): return '' if self.label is None else self.label + ': '
    # def getLabelPrefix(self): return self.getLabel() + ': '

    # For debugging
    def __str__(self):   # pragma: no cover
        if isinstance(self, StatementBlock):
            return 'Sb'+str(self.id)
        return type(self).__name__[:3]+str(self.id)
    __repr__ = __str__

class TryStatement(LazyLabelBase):
    def __init__(self, labelfunc, begink, endk, tryb, pairs):
        super(TryStatement, self).__init__(labelfunc, begink, endk)
        self.tryb, self.pairs = tryb, pairs

    def getScopes(self): return (self.tryb,) + zip(*self.pairs)[1]

    def print_(self, printer, print_):
        tryb = print_(self.tryb)
        parts = ['catch({}) {}'.format(print_(x), print_(y)) for x,y in self.pairs]
        return '{}try {} {}'.format(self.getLabelPrefix(), tryb, '\n'.join(parts))

    def tree(self, printer, tree):
        parts = [map(tree, t) for t in self.pairs]
        return [self.__class__.__name__, self.label, tree(self.tryb), parts]

class IfStatement(LazyLabelBase):
    def __init__(self, labelfunc, begink, endk, expr, scopes):
        super(IfStatement, self).__init__(labelfunc, begink, endk)
        self.expr = expr # don't rename without changing how var replacement works!
        self.scopes = scopes

    def getScopes(self): return self.scopes

    def print_(self, printer, print_):
        lbl = self.getLabelPrefix()
        parts = [self.expr] + list(self.scopes)

        if len(self.scopes) == 1:
            parts = [print_(x) for x in parts]
            return '{}if ({}) {}'.format(lbl, *parts)

        # Special case handling for 'else if'
        fblock = self.scopes[1]
        if len(fblock.statements) == 1:
            stmt = fblock.statements[-1]
            if isinstance(stmt, IfStatement) and stmt.label is None:
                parts[-1] = stmt
        parts = [print_(x) for x in parts]
        return '{}if ({}) {} else {}'.format(lbl, *parts)

    def tree(self, printer, tree): return [self.__class__.__name__, self.label, tree(self.expr), map(tree, self.scopes)]

class SwitchStatement(LazyLabelBase):
    def __init__(self, labelfunc, begink, endk, expr, pairs):
        super(SwitchStatement, self).__init__(labelfunc, begink, endk)
        self.expr = expr # don't rename without changing how var replacement works!
        self.pairs = pairs

    def getScopes(self): return zip(*self.pairs)[1]
    def hasDefault(self): return None in zip(*self.pairs)[0]

    def print_(self, printer, print_):
        expr = print_(self.expr)

        def printCase(keys):
            if keys is None:
                return 'default: '
            assert keys
            return ''.join(map('case {}: '.format, sorted(keys)))

        bodies = [(printCase(keys) + print_(scope)) for keys, scope in self.pairs]
        if self.pairs[-1][0] is None and len(self.pairs[-1][1].statements) == 0:
            bodies.pop()

        contents = '\n'.join(bodies)
        indented = ['    '+line for line in contents.splitlines()]
        lines = ['{'] + indented + ['}']
        return '{}switch({}) {}'.format(self.getLabelPrefix(), expr, '\n'.join(lines))

    def tree(self, printer, tree):
        parts = []
        for keys, scope in self.pairs:
            parts.append([[None] if keys is None else sorted(keys), tree(scope)])
        return [self.__class__.__name__, self.label, tree(self.expr), parts]

class WhileStatement(LazyLabelBase):
    def __init__(self, labelfunc, begink, endk, parts):
        super(WhileStatement, self).__init__(labelfunc, begink, endk)
        self.expr = Literal.TRUE
        self.parts = parts
        assert len(self.parts) == 1

    def getScopes(self): return self.parts

    def print_(self, printer, print_):
        parts = print_(self.expr), print_(self.parts[0])
        return '{}while({}) {}'.format(self.getLabelPrefix(), *parts)

    def tree(self, printer, tree): return [self.__class__.__name__, self.label, tree(self.expr), tree(self.parts[0])]

class StatementBlock(LazyLabelBase):
    def __init__(self, labelfunc, begink, endk, statements, jumpk, labelable=True):
        super(StatementBlock, self).__init__(labelfunc, begink, endk)
        self.parent = None # should be assigned later
        self.statements = statements
        self.jumpKey = jumpk
        self.labelable = labelable

    def doesFallthrough(self): return self.jumpKey is None or self.jumpKey == self.breakKey

    def getScopes(self): return self,

    def print_(self, printer, print_):
        assert self.labelable or self.label is None
        contents = '\n'.join(print_(x) for x in self.statements)
        indented = ['    '+line for line in contents.splitlines()]
        # indented[:0] = ['    //{} {}'.format(self,x) for x in (self.continueKey, self.breakKey, self.jumpKey)]
        lines = [self.getLabelPrefix() + '{'] + indented + ['}']
        return '\n'.join(lines)

    @staticmethod
    def join(*scopes):
        blists = [s.bases for s in scopes if s is not None] # allow None to represent the universe (top element)
        if not blists:
            return None
        common = [x for x in zip(*blists) if len(set(x)) == 1]
        return common[-1][0]

    def tree(self, printer, tree): return ['BlockStatement', self.label, map(tree, self.statements)]

#############################################################################################################################################
# Careful, order is important here!
_assignable_sprims = objtypes.ByteTT, objtypes.ShortTT, objtypes.CharTT
_assignable_lprims = objtypes.IntTT, objtypes.LongTT, objtypes.FloatTT, objtypes.DoubleTT

# Also used in boolize.py
def isPrimativeAssignable(x, y): # x = fromt, y = to
    assert objtypes.dim(x) == objtypes.dim(y) == 0

    if x == y or (x in _assignable_sprims and y in _assignable_lprims):
        return True
    elif (x in _assignable_lprims and y in _assignable_lprims):
        return _assignable_lprims.index(x) <= _assignable_lprims.index(y)
    else:
        return (x, y) == (objtypes.ByteTT, objtypes.ShortTT)

def isReferenceType(tt):
    return tt == objtypes.NullTT or objtypes.dim(tt) or (objtypes.className(tt) is not None)

def isJavaAssignable(env, fromt, to):
    if fromt is None or to is None: # this should never happen, except during debugging
        return True

    if isReferenceType(to):
        assert isReferenceType(fromt)
        # todo - make it check interfaces too
        return objtypes.isSubtype(env, fromt, to)
    else: # allowed if numeric conversion is widening
        return isPrimativeAssignable(fromt, to)

_int_tts = objtypes.LongTT, objtypes.IntTT, objtypes.ShortTT, objtypes.CharTT, objtypes.ByteTT
def makeCastExpr(newtt, expr, fixEnv=None):
    if newtt == expr.dtype:
        return expr

    # if casting a literal with compatible type, just create a literal of the new type
    if isinstance(expr, Literal):
        allowed_conversions = [
            (objtypes.FloatTT, objtypes.DoubleTT),
            (objtypes.IntTT, objtypes.LongTT),
            (objtypes.IntTT, objtypes.BoolTT),
            (objtypes.BoolTT, objtypes.IntTT),
        ]
        if (expr.dtype, newtt) in allowed_conversions:
            return Literal(newtt, expr.val)

    if newtt == objtypes.IntTT and expr.dtype == objtypes.BoolTT:
        return Ternary(expr, Literal.ONE, Literal.ZERO)
    elif newtt == objtypes.BoolTT and expr.dtype == objtypes.IntTT:
        return BinaryInfix('!=', [expr, Literal.ZERO], objtypes.BoolTT)

    ret = Cast(TypeName(newtt), expr)
    if fixEnv is not None:
        ret = ret.fix(fixEnv)
    return ret
#############################################################################################################################################
# Precedence:
#    0 - pseudoprimary
#    5 - pseudounary
#    10-19 binary infix
#    20 - ternary
#    21 - assignment
# Associativity: L = Left, R = Right, A = Full

class JavaExpression(object):
    precedence = 0 # Default precedence
    params = [] # for subclasses that don't have params

    def complexity(self): return 1 + max(e.complexity() for e in self.params) if self.params else 0

    def postFlatIter(self):
        return itertools.chain([self], *[expr.postFlatIter() for expr in self.params])

    def print_(self, printer, print_):
        return self.fmt.format(*[print_(expr) for expr in self.params])

    def tree(self, printer, tree): return [self.__class__.__name__, map(tree, self.params)]

    def replaceSubExprs(self, rdict):
        if self in rdict:
            return rdict[self]
        self.params = [param.replaceSubExprs(rdict) for param in self.params]
        return self

    def fixLiterals(self):
        self.params = [param.fixLiterals() for param in self.params]
        return self

    def addCasts(self, env):
        for param in self.params:
            param.addCasts(env)
        self.addCasts_sub(env)

    def addCasts_sub(self, env): pass

    def addParens(self):
        for param in self.params:
            param.addParens()
        self.params = list(self.params) # Copy before editing, just to be extra safe
        self.addParens_sub()

    def addParens_sub(self): pass

    def isLocalAssign(self): return isinstance(self, Assignment) and isinstance(self.params[0], Local)

    def __repr__(self):   # pragma: no cover
        return type(self).__name__.rpartition('.')[-1] + ' ' + visitor.DefaultVisitor().visit(self)
    __str__ = __repr__

class ArrayAccess(JavaExpression):
    def __init__(self, *params):
        if params[0].dtype == objtypes.NullTT:
            # Unfortunately, Java doesn't really support array access on null constants
            #So we'll just cast it to Object[] as a hack
            param = makeCastExpr(objtypes.withDimInc(objtypes.ObjectTT, 1), params[0])
            params = param, params[1]

        self.params = list(params)
        self.fmt = '{}[{}]'

    @property
    def dtype(self): return objtypes.withDimInc(self.params[0].dtype, -1)

    def addParens_sub(self):
        p0 = self.params[0]
        if p0.precedence > 0 or isinstance(p0, ArrayCreation):
            self.params[0] = Parenthesis(p0)

class ArrayCreation(JavaExpression):
    def __init__(self, tt, *sizeargs):
        self.dim = objtypes.dim(tt)
        self.params = [TypeName(objtypes.withNoDim(tt))] + list(sizeargs)
        self.dtype = tt
        assert self.dim >= len(sizeargs) > 0
        self.fmt = 'new {}' + '[{}]'*len(sizeargs) + '[]'*(self.dim-len(sizeargs))

    def tree(self, printer, tree): return [self.__class__.__name__, map(tree, self.params), self.dim]

class Assignment(JavaExpression):
    precedence = 21
    def __init__(self, *params):
        self.params = list(params)
        self.fmt = '{} = {}'

    @property
    def dtype(self): return self.params[0].dtype

    def addCasts_sub(self, env):
        left, right = self.params
        if not isJavaAssignable(env, right.dtype, left.dtype):
            expr = makeCastExpr(left.dtype, right, fixEnv=env)
            self.params = [left, expr]

    def tree(self, printer, tree): return [self.__class__.__name__, map(tree, self.params), '']

_binary_ptable = ['* / %', '+ -', '<< >> >>>',
    '< > <= >= instanceof', '== !=',
    '&', '^', '|', '&&', '||']

binary_precedences = {}
for _ops, _val in zip(_binary_ptable, range(10,20)):
    for _op in _ops.split():
        binary_precedences[_op] = _val

class BinaryInfix(JavaExpression):
    def __init__(self, opstr, params, dtype=None):
        assert len(params) == 2
        self.params = params
        self.opstr = opstr
        self.fmt = '{{}} {} {{}}'.format(opstr)
        self._dtype = dtype
        self.precedence = binary_precedences[opstr]

    @property
    def dtype(self): return self.params[0].dtype if self._dtype is None else self._dtype

    def addParens_sub(self):
        myprec = self.precedence
        associative = myprec >= 15 # for now we treat +, *, etc as nonassociative due to floats

        for i, p in enumerate(self.params):
            if p.precedence > myprec:
                self.params[i] = Parenthesis(p)
            elif p.precedence == myprec and i > 0 and not associative:
                self.params[i] = Parenthesis(p)

    def tree(self, printer, tree): return [self.__class__.__name__, map(tree, self.params), self.opstr]

class Cast(JavaExpression):
    precedence = 5
    def __init__(self, *params):
        self.dtype = params[0].tt
        self.params = list(params)
        self.fmt = '({}){}'

    def fix(self, env):
        tt, expr = self.dtype, self.params[1]
        # "Impossible" casts are a compile error in Java.
        # This can be fixed with an intermediate cast to Object
        if isReferenceType(tt):
            if not isJavaAssignable(env, tt, expr.dtype):
                if not isJavaAssignable(env, expr.dtype, tt):
                    expr = makeCastExpr(objtypes.ObjectTT, expr)
                    self.params = [self.params[0], expr]
        return self

    def addCasts_sub(self, env): self.fix(env)
    def addParens_sub(self):
        p1 = self.params[1]
        if p1.precedence > 5 or (isinstance(p1, UnaryPrefix) and p1.opstr[0] in '-+'):
            self.params[1] = Parenthesis(p1)

class ClassInstanceCreation(JavaExpression):
    def __init__(self, typename, tts, arguments):
        self.typename, self.tts, self.params = typename, tts, arguments
        self.dtype = typename.tt

    def print_(self, printer, print_):
        return 'new {}({})'.format(print_(self.typename), ', '.join(print_(x) for x in self.params))

    def tree(self, printer, tree):
        return [self.__class__.__name__, map(tree, self.params), tree(self.typename)]

    def addCasts_sub(self, env):
        newparams = []
        for tt, expr in zip(self.tts, self.params):
            if expr.dtype != tt and (ALWAYS_CAST_PARAMS or not isJavaAssignable(env, expr.dtype, tt)):
                expr = makeCastExpr(tt, expr, fixEnv=env)
            newparams.append(expr)
        self.params = newparams

class FieldAccess(JavaExpression):
    def __init__(self, primary, name, dtype, op=None, printLeft=True):
        self.dtype = dtype
        self.params = [primary]
        self.op, self.name = op, name
        self.printLeft = printLeft
        # self.params, self.name = [primary], escapeString(name)
        # self.fmt = ('{}.' if printLeft else '') + self.name

    def print_(self, printer, print_):
        if self.op is None:
            name = self.name
            assert name in ('length','class')
        else:
            cls, name, desc = self.op.target, self.op.name, self.op.desc
            name = escapeString(printer.fieldName(cls, name, desc))
        pre = print_(self.params[0])+'.' if self.printLeft else ''
        return pre+name

    def tree(self, printer, tree):
        if self.op is None:
            trip = None, self.name, None
        else:
            trip = self.op.target, self.op.name, self.op.desc
        return [self.__class__.__name__, map(tree, self.params), trip, self.printLeft]

    def addParens_sub(self):
        p0 = self.params[0]
        if p0.precedence > 0:
            self.params[0] = Parenthesis(p0)

def printFloat(x, isSingle):
    assert x >= 0.0 and not math.isinf(x)
    suffix = 'f' if isSingle else ''
    if isSingle and x > 0.0:
        # Try to find more compract representation for floats, since repr treats everything as doubles
        m, e = math.frexp(x)
        half_ulp2 = math.ldexp(1.0, max(e - 25, -150)) # don't bother doubling when near the upper range of a given e value
        half_ulp1 = (half_ulp2/2) if m == 0.5 and e >= -125 else half_ulp2
        lbound, ubound = x-half_ulp1, x+half_ulp2
        assert lbound < x < ubound
        s = '{:g}'.format(x).replace('+','')
        if lbound < float(s) < ubound: # strict ineq to avoid potential double rounding issues
            return s + suffix
    return repr(x) + suffix

class Literal(JavaExpression):
    def __init__(self, vartype, val):
        self.dtype = vartype
        self.val = val
        if self.dtype == objtypes.ClassTT:
            self.params = [TypeName(val)]

    def getStr(self):
        if self.dtype == objtypes.StringTT:
            return '"' + escapeString(self.val) + '"'
        elif self.dtype == objtypes.IntTT:
            return str(self.val)
        elif self.dtype == objtypes.LongTT:
            return str(self.val) + 'L'
        elif self.dtype == objtypes.FloatTT or self.dtype == objtypes.DoubleTT:
            return printFloat(self.val, self.dtype == objtypes.FloatTT)
        elif self.dtype == objtypes.NullTT:
            return 'null'
        elif self.dtype == objtypes.BoolTT:
            return 'true' if self.val else 'false'

    def fixLiterals(self):
        # From the point of view of the Java Language, there is no such thing as a negative literal.
        # This replaces invalid literal values with unary minus (and division for non-finite floats)
        if self.dtype == objtypes.IntTT or self.dtype == objtypes.LongTT:
            if self.val < 0:
                return UnaryPrefix('-', Literal(self.dtype, -self.val))
        elif self.dtype == objtypes.FloatTT or self.dtype == objtypes.DoubleTT:
            x = self.val
            zero = Literal.DZERO if self.dtype == objtypes.DoubleTT else Literal.FZERO
            if math.isnan(x):
                return BinaryInfix('/', [zero, zero])
            elif math.isinf(x): #+/- inf
                numerator = Literal(self.dtype, math.copysign(1.0, x)).fixLiterals()
                return BinaryInfix('/', [numerator, zero])
            # finite negative numbers
            if math.copysign(1.0, x) == -1.0:
                return UnaryPrefix('-', Literal(self.dtype, math.copysign(x, 1.0)))

        return self

    def print_(self, printer, print_):
        if self.dtype == objtypes.ClassTT:
            # for printing class literals
            return '{}.class'.format(print_(self.params[0]))
        return self.getStr()

    def tree(self, printer, tree):
        result = tree(self.params[0]) if self.dtype == objtypes.ClassTT else self.getStr()
        return [self.__class__.__name__, result, self.dtype]

    def _key(self): return self.dtype, self.val
    def __eq__(self, other): return type(self) == type(other) and self._key() == other._key()
    def __ne__(self, other): return type(self) != type(other) or self._key() != other._key()
    def __hash__(self): return hash(self._key())
Literal.FALSE = Literal(objtypes.BoolTT, 0)
Literal.TRUE = Literal(objtypes.BoolTT, 1)
Literal.N_ONE = Literal(objtypes.IntTT, -1)
Literal.ZERO = Literal(objtypes.IntTT, 0)
Literal.ONE = Literal(objtypes.IntTT, 1)

Literal.LZERO = Literal(objtypes.LongTT, 0)
Literal.FZERO = Literal(objtypes.FloatTT, 0.0)
Literal.DZERO = Literal(objtypes.DoubleTT, 0.0)
Literal.NULL = Literal(objtypes.NullTT, None)

_init_d = {objtypes.BoolTT: Literal.FALSE,
        objtypes.IntTT: Literal.ZERO,
        objtypes.LongTT: Literal.LZERO,
        objtypes.FloatTT: Literal.FZERO,
        objtypes.DoubleTT: Literal.DZERO}
def dummyLiteral(tt):
    return _init_d.get(tt, Literal.NULL)

class Local(JavaExpression):
    def __init__(self, vartype, namefunc):
        self.dtype = vartype
        self.name = None
        self.func = namefunc

    def print_(self, printer, print_):
        if self.name is None:
            self.name = self.func(self)
        return self.name

    def tree(self, printer, tree): return [self.__class__.__name__, self.print_(None, None)]

class MethodInvocation(JavaExpression):
    def __init__(self, left, name, tts, arguments, op, dtype):
        if left is None:
            self.params = arguments
        else:
            self.params = [left] + arguments
        self.hasLeft = (left is not None)
        self.dtype = dtype
        self.name = name
        self.tts = tts
        self.op = op # keep around for future reference and new merging

    def print_(self, printer, print_):
        cls, name, desc = self.op.target, self.op.name, self.op.desc
        if name != self.name:
            assert name == '<init>'
            name = self.name
        else:
            name = escapeString(printer.methodName(cls, name, desc))

        if self.hasLeft:
            left, arguments = self.params[0], self.params[1:]
            return '{}.{}({})'.format(print_(left), name, ', '.join(print_(x) for x in arguments))
        else:
            arguments = self.params
            return '{}({})'.format(name, ', '.join(print_(x) for x in arguments))

    def tree(self, printer, tree):
        trip = self.op.target, self.op.name, self.op.desc
        return [self.__class__.__name__, map(tree, self.params), trip, self.name, self.hasLeft]

    def addCasts_sub(self, env):
        newparams = []
        for tt, expr in zip(self.tts, self.params):
            if expr.dtype != tt and (ALWAYS_CAST_PARAMS or not isJavaAssignable(env, expr.dtype, tt)):
                expr = makeCastExpr(tt, expr, fixEnv=env)
            newparams.append(expr)
        self.params = newparams

    def addParens_sub(self):
        if self.hasLeft:
            p0 = self.params[0]
            if p0.precedence > 0:
                self.params[0] = Parenthesis(p0)

class Parenthesis(JavaExpression):
    def __init__(self, param):
        self.params = [param]
        self.fmt = '({})'

    @property
    def dtype(self): return self.params[0].dtype

class Ternary(JavaExpression):
    precedence = 20
    def __init__(self, *params):
        self.params = list(params)
        self.fmt = '{} ? {} : {}'

    @property
    def dtype(self): return self.params[1].dtype

    def addParens_sub(self):
        # Add unecessary parenthesis to complex conditions for readability
        if self.params[0].precedence >= 20 or self.params[0].complexity() > 0:
            self.params[0] = Parenthesis(self.params[0])
        if self.params[2].precedence > 20:
            self.params[2] = Parenthesis(self.params[2])

class TypeName(JavaExpression):
    def __init__(self, tt):
        self.dtype = None
        self.tt = tt

    def print_(self, printer, print_):
        name = objtypes.className(self.tt)
        if name is not None:
            name = printer.className(name)
            name = escapeString(name.replace('/','.'))
            if name.rpartition('.')[0] == 'java.lang':
                name = name.rpartition('.')[2]
        else:
            name = objtypes.primName(self.tt)
        s = name + '[]'*objtypes.dim(self.tt)
        return s

    def tree(self, printer, tree): return [self.__class__.__name__, self.tt]

    def complexity(self): return -1 # exprs which have this as a param won't be bumped up to 1 uncessarily

class CatchTypeNames(JavaExpression): # Used for caught exceptions, which can have multiple types specified
    def __init__(self, env, tts):
        assert(tts and not any(objtypes.dim(tt) for tt in tts)) # at least one type, no array types
        self.tnames = map(TypeName, tts)
        self.dtype = objtypes.commonSupertype(env, tts)

    def print_(self, printer, print_):
        return ' | '.join(print_(tn) for tn in self.tnames)

    def tree(self, printer, tree): return [self.__class__.__name__, map(tree, self.tnames)]

class UnaryPrefix(JavaExpression):
    precedence = 5
    def __init__(self, opstr, param, dtype=None):
        self.params = [param]
        self.opstr = opstr
        self.fmt = opstr + '{}'
        self._dtype = dtype

    @property
    def dtype(self): return self.params[0].dtype if self._dtype is None else self._dtype

    def addParens_sub(self):
        p0 = self.params[0]
        if p0.precedence > 5 or (isinstance(p0, UnaryPrefix) and p0.opstr[0] == self.opstr[0]):
            self.params[0] = Parenthesis(p0)

    def tree(self, printer, tree): return ['Unary', map(tree, self.params), self.opstr, False]

class Dummy(JavaExpression):
    def __init__(self, fmt, params, isNew=False, dtype=None):
        self.params = params
        self.fmt = fmt
        self.isNew = isNew
        self.dtype = dtype



class ClassLoaderError(Exception):
    def __init__(self, typen=None, data=""):
        self.type = typen
        self.data = data

        message = "\n{}: {}".format(typen, data) if typen else data
        super(ClassLoaderError, self).__init__(message)

class VerificationError(Exception):
    def __init__(self, message, data=None):
        super(VerificationError, self).__init__(message)
        self.data = data
